"""
This module defines variables that depend on Timeweb Cloud products.
They are subject to change at any time as cloud computing capabilities
expand or other infrastructure or product changes occur.
"""

# Service URLs
CONTROL_PANEL_URL = "https://timeweb.cloud/my"

# Location specific parameters. May change later.
REGIONS_WITH_IPV6 = ["ru-1", "ru-3", "pl-1", "nl-1", "kz-1"]
REGIONS_WITH_IMAGES = ["ru-1", "ru-3", "kz-1", "pl-1", "nl-1"]
REGIONS_WITH_LAN = ["ru-1", "ru-3", "nl-1", "pl-1", "de-1"]
ZONES_WITH_LAN = [
    "spb-1",
    "spb-3",
    "spb-4",
    "msk-1",
    "ams-1",
    "gdn-1",
    "fra-1",
]
# The default availability zones per regions.
REGION_ZONE_MAP = {
    "ru-1": "spb-3",
    "ru-2": "nsk-1",
    "ru-3": "msk-1",
    "kz-1": "ala-1",
    "pl-1": "gdn-1",
    "nl-1": "ams-1",
    "de-1": "fra-1",
}



"""Commands."""

from .account import account, whoami
from .apps import apps
from .config import config
from .server import server
from .ssh_key import ssh_key
from .image import image
from .project import project
from .database import database
from .storage import storage
from .balancer import balancer
from .kubernetes import cluster
from .domain import domain
from .vpc import vpc
from .firewall import firewall
from .floating_ip import floating_ip



"""CLI for Timeweb Cloud services."""

from typing import Optional
from pathlib import Path

import typer

from .typerx import TyperAlias
from .commands import (
    config,
    account,
    server,
    ssh_key,
    image,
    project,
    database,
    storage,
    balancer,
    cluster,
    domain,
    vpc,
    firewall,
    floating_ip,
    whoami,
    apps,
)
from .commands.common import version_callback, version_option, verbose_option


cli = TyperAlias(
    help=__doc__,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
cli.add_typer(config, name="config")
cli.add_typer(account, name="account")
cli.add_typer(apps, name="apps")
cli.add_typer(server, name="server", aliases=["servers", "s"])
cli.add_typer(ssh_key, name="ssh-key", aliases=["ssh-keys", "k"])
cli.add_typer(image, name="image", aliases=["images", "i"])
cli.add_typer(project, name="project", aliases=["projects", "p"])
cli.add_typer(database, name="database", aliases=["databases", "db"])
cli.add_typer(storage, name="storage", aliases=["storages", "s3"])
cli.add_typer(balancer, name="balancer", aliases=["balancers", "lb"])
cli.add_typer(
    cluster, name="cluster", aliases=["clusters", "kubernetes", "k8s"]
)
cli.add_typer(domain, name="domain", aliases=["domains", "d"])
cli.add_typer(vpc, name="vpc", aliases=["vpcs", "network", "networks"])
cli.add_typer(firewall, name="firewall", aliases=["fw"])
cli.add_typer(floating_ip, name="ip", aliases=["ips"])
cli.add_typer(whoami, name="whoami", aliases=[])


@cli.command("version")
def version_command():
    """Show version and exit."""
    version_callback(True)


@cli.callback()
def root(
    version: Optional[bool] = version_option,
    verbose: Optional[bool] = verbose_option,
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        envvar="TWC_CONFIG_FILE",
        show_envvar=False,
        show_default=False,
        exists=True,
        dir_okay=False,
        help="Use config.",
        rich_help_panel="Global options",
    ),
    profile: Optional[str] = typer.Option(
        None,
        "--profile",
        "-p",
        metavar="NAME",
        envvar="TWC_PROFILE",
        show_envvar=False,
        show_default=False,
        help="Use profile.",
        rich_help_panel="Global options",
    ),
):
    # pylint: disable=unused-argument
    """Callback for root level command options."""


if __name__ == "__main__":
    cli()



"""Base module for Timeweb Cloud API client."""

import logging
import textwrap
from typing import Optional, Callable

import requests

from twc.__version__ import __version__, __pyversion__
from . import exceptions as exc


class TimewebCloudBase:
    """Base class for Timeweb Cloud API client."""

    # pylint: disable=method-hidden
    # pylint: disable=invalid-name

    API_BASE_URL = "https://api.timeweb.cloud"
    TIMEOUT = 100
    USER_AGENT = f"TWC-CLI/{__version__} Python {__pyversion__}"

    def __init__(
        self,
        api_token: str,
        api_base_url: Optional[str] = API_BASE_URL,
        api_path: Optional[str] = "/api/v1",
        headers: Optional[dict] = None,
        user_agent: Optional[str] = USER_AGENT,
        timeout: Optional[int] = TIMEOUT,
        hide_token: Optional[bool] = True,
        request_decorator: Optional[Callable] = None,
    ):
        self.api_token = api_token
        self.api_base_url = api_base_url
        self.api_path = api_path
        self.api_url = self.api_base_url + self.api_path
        self.api_url_v1 = self.api_base_url + "/api/v1"
        self.api_url_v2 = self.api_base_url + "/api/v2"
        self.timeout = timeout
        self.headers = requests.utils.default_headers()
        self.headers["User-Agent"] = user_agent
        self.headers["Authorization"] = f"Bearer {self.api_token}"
        self.log = logging.getLogger("api_client")
        self.hide_token = hide_token

        if headers:
            self.headers.update(headers)

        # Decorate _request()
        if request_decorator is not None:
            self._request = request_decorator(self._request)

    def _format_headers(self, headers: dict) -> str:
        """Format HTTP headers for log."""
        return "\n".join(f"{k}: {v}" for k, v in headers.items())

    def _secure_log(self, headers: dict) -> dict:
        """Replace API access token with placeholder."""
        _headers = headers.copy()
        if self.hide_token:
            _headers["Authorization"] = "Bearer <SENSITIVE_DATA_DELETED>"
        return _headers

    def _log_request(self, response: requests.Response) -> None:
        """Log HTTP requests."""
        res_body = response.text or "<NO_BODY>"
        req_body = response.request.body or "<NO_BODY>"
        if isinstance(req_body, (bytes, bytearray)):
            req_body = req_body.decode()

        self.log.debug(
            textwrap.dedent(  # pylint: disable=logging-format-interpolation
                """
            ---------------- Request ----------------
            {req.method} {req.url}
            {req_headers}

            {req_body}
            ---------------- Response ---------------
            {res.status_code} {res.reason} {res.url}
            {res_headers}

            {res_body}"""
            ).format(
                req=response.request,
                req_headers=self._format_headers(
                    self._secure_log(response.request.headers)
                ),
                req_body=req_body,
                res=response,
                res_headers=self._format_headers(response.headers),
                res_body=res_body,
            )
        )

    def _request(
        self,
        method: str,
        url: str,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        files: Optional[dict] = None,
        timeout: Optional[int] = None,
    ) -> requests.Response:
        """Make request and handle errors."""

        if not timeout:
            timeout = self.timeout

        if not headers:
            headers = self.headers

        _headers = self._secure_log(headers)
        self.log.debug("Called with args: %s %s %s", method, url, _headers)

        try:
            response = requests.request(
                method,
                url,
                headers=headers,
                params=params,
                data=data,
                json=json,
                files=files,
                timeout=timeout,
            )
        except requests.exceptions.ConnectionError as conerr:
            raise exc.NetworkError(f"Coul'd not connect to server: {conerr}")

        try:
            response.raise_for_status()
            self._log_request(response)
        except requests.HTTPError as e:
            self._log_request(response)

            # API issue: Bad response: 401 Unauthorized response haven't body
            # There is workaround about it - raise UnauthorizedError before
            # checking a JSON schema.
            if response.status_code == 401:
                raise exc.UnauthorizedError(
                    request=e.request,
                    response=e.response,
                ) from e

            try:
                error = exc.ErrResponse(**response.json())
            except requests.JSONDecodeError as err:
                raise exc.MalformedResponseError(
                    message="Response have no JSON schema or have invalid JSON syntax.",
                    request=e.request,
                    response=e.response,
                ) from err

            if response.status_code == 400:
                raise exc.BadRequestError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            if response.status_code == 403:
                raise exc.ForbiddenError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            if response.status_code == 404:
                raise exc.NotFoundError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            if response.status_code == 409:
                raise exc.ConflictError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            if response.status_code == 423:
                raise exc.LockedError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            if response.status_code == 429:
                raise exc.TooManyRequestsError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            if response.status_code == 500:
                raise exc.InternalServerError(
                    request=e.request,
                    response=e.response,
                    status_code=error.status_code,
                    error_code=error.error_code,
                    message=error.message,
                    response_id=error.response_id,
                ) from e
            raise exc.UnexpectedResponseError(e) from e

        return response



"""Manage disk images."""

import re
import sys
from logging import debug
from typing import Optional, List
from pathlib import Path
from uuid import UUID

import typer
from click import UsageError
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.apiwrap import create_client
from twc.api import ServerOSType
from .common import (
    verbose_option,
    config_option,
    profile_option,
    filter_option,
    yes_option,
    output_format_option,
    region_option,
)


image = TyperAlias(help=__doc__)


# ------------------------------------------------------------- #
# $ twc image list                                              #
# ------------------------------------------------------------- #


def print_images(response: Response, filters: Optional[str] = None):
    """Print table with images list."""
    images = response.json()["images"]
    if filters:
        images = fmt.filter_list(images, filters)

    table = fmt.Table()
    table.header(
        [
            "ID",
            # "NAME",
            "TYPE",
            "REGION",
            "STATUS",
            "SIZE",
        ]
    )
    for img in images:
        table.row(
            [
                img["id"],
                img["type"],
                # img["name"][:18] + "..."
                # if len(img["name"]) > 18
                # else img["name"],
                img["location"],
                img["status"],
                str(round(img["size"] / 1024)) + "G",
            ]
        )
    table.print()


@image.command("list", "ls")
def image_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    limit: int = typer.Option(500, help="Items to display."),
    filters: Optional[str] = filter_option,
):
    """List images."""
    client = create_client(config, profile)
    response = client.get_images(limit=limit)
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_images,
    )


# ------------------------------------------------------------- #
# $ twc image get                                               #
# ------------------------------------------------------------- #


def print_image(response: Response):
    """Print table with image info."""
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "TYPE",
            "REGION",
            "STATUS",
            "DISK",
            "SIZE",
        ]
    )
    img = response.json()["image"]
    table.row(
        [
            img["id"],
            img["name"],
            img["type"],
            img["location"],
            img["status"],
            img["disk_id"],
            str(round(img["size"] / 1024)) + "G",
        ]
    )
    table.print()


@image.command("get")
def image_get(
    image_id: UUID,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    status: Optional[bool] = typer.Option(
        False,
        "--status",
        help="Display status and exit with 0 if status is 'created'.",
    ),
):
    """Get image info."""
    client = create_client(config, profile)
    response = client.get_image(image_id)
    if status:
        state = response.json()["image"]["status"]
        if state == "created":
            print(state)
            raise typer.Exit()
        sys.exit(state)
    fmt.printer(response, output_format=output_format, func=print_image)


# ------------------------------------------------------------- #
# $ twc image create                                            #
# ------------------------------------------------------------- #


@image.command("create")
def image_create(
    disk_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(
        None, help="Image human readable name."
    ),
    desc: Optional[str] = typer.Option(None, help="Image description."),
):
    """Create image from existing Cloud Server disk."""
    client = create_client(config, profile)
    response = client.create_image(
        disk_id=disk_id, name=name, description=desc
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["image"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc image remove                                            #
# ------------------------------------------------------------- #


@image.command("remove", "rm")
def image_remove(
    image_ids: List[UUID] = typer.Argument(..., metavar="IMAGE_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove image."""
    if not yes:
        typer.confirm(
            "This action cannot be undone. Continue?",
            abort=True,
        )
    client = create_client(config, profile)
    for image_id in image_ids:
        response = client.delete_image(image_id)
        if response.status_code == 204:
            print(image_id)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc image set                                               #
# ------------------------------------------------------------- #


@image.command("set")
def image_set_property(
    image_id: UUID,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(
        None, help="Image human readable name."
    ),
    desc: Optional[str] = typer.Option(None, help="Image description."),
):
    """Change image name and description."""
    if not name and not desc:
        raise UsageError(
            "Nothing to do. Set one of options: ['--name', '--desc']"
        )
    client = create_client(config, profile)
    response = client.update_image(image_id, name=name, description=desc)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["image"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc image upload                                            #
# ------------------------------------------------------------- #


@image.command("upload")
def image_upload(
    file: str = typer.Argument(..., help="Direct HTTP(S) link to image."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(
        None, help="Image human readable name."
    ),
    desc: Optional[str] = typer.Option(None, help="Image description."),
    os_type: ServerOSType = typer.Option(
        ServerOSType.OTHER.value,
        metavar="OS_TYPE",
        case_sensitive=False,
        help="OS type. This value is formal and not affects on server/image.",
    ),
    region: Optional[str] = region_option,
):
    """Upload image from URL."""
    client = create_client(config, profile)
    if re.match(r"https?://", file):
        debug(f"Upload URL: {file}")
    else:
        sys.exit(f"Invalid link: {file}")

    response = client.create_image(
        upload_url=file,
        name=name,
        description=desc,
        os_type=os_type,
        location=region,
    )

    # FUTURE: Implement file upload from local disk

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["image"]["id"]),
    )



"""Manage CLI configuration."""

import re
import os
import sys
import ctypes
import json
from typing import Optional, List
from enum import Enum
from pathlib import Path

import typer
import toml
import yaml

from twc.utils import merge_dicts
from twc.fmt import print_colored
from .common import (
    default_config_file,
    load_config,
    config_callback,
    config_option,
    profile_option,
)


config = typer.Typer(help=__doc__, no_args_is_help=False)


DEFAULT_CONFIG = {"default": {"token": None}}


def write_to_file(data: dict, filepath: Path) -> None:
    """Dump `data` dict into TOML and write into file."""
    try:
        with open(filepath, "w", encoding="utf-8") as file:
            toml.dump(data, file)
        if os.name == "nt":
            hidden_file_attr = 0x02
            ctypes.windll.kernel32.SetFileAttributesW(
                str(filepath), hidden_file_attr
            )
        print(f"Done! Configuration is saved in {filepath}")
        sys.exit(0)
    except OSError as error:
        sys.exit(f"Error: Cannot write configration file: {error}")


def make_config(filepath: Path = default_config_file()) -> None:
    """Create new configuration file or edit existing profile token."""
    # Edit existing file
    if os.path.exists(filepath):
        if typer.confirm(
            "You already have TWC CLI configured, continue?",
            default=False,
        ):
            current_config = load_config()
            profile = typer.prompt("Enter profile name", default="default")
            token = typer.prompt(f"Enter API token for '{profile}'")
            try:
                # Edit existing profile
                current_config[profile] = merge_dicts(
                    current_config[profile],
                    {"token": token},
                )
            except KeyError:
                # Add new profile
                current_config[profile] = {}  # init new config section
                current_config[profile] = merge_dicts(
                    current_config[profile],
                    {"token": token},
                )
            write_to_file(current_config, filepath)
        else:
            sys.exit("Aborted!")
    # Make new file
    print("Create new configuration file. Enter your API token.")
    while True:
        token = input("Token: ")
        if token:
            DEFAULT_CONFIG.update({"default": {"token": token}})
            break
        print("Please enter token. Press ^C to cancel.")
    write_to_file(DEFAULT_CONFIG, filepath)


# ------------------------------------------------------------- #
# $ twc config                                                  #
# ------------------------------------------------------------- #


@config.callback(invoke_without_command=True)
def config_root_cmd(ctx: typer.Context):
    """Make new configuration file if not exist or edit existing."""
    if ctx.invoked_subcommand is None:
        make_config()


# ------------------------------------------------------------- #
# $ twc config init                                             #
# ------------------------------------------------------------- #


@config.command("init")
def config_init():
    """Make new configuration file if not exist or edit existing."""
    make_config()


# ------------------------------------------------------------- #
# $ twc config file                                             #
# ------------------------------------------------------------- #


@config.command("file")
def config_file(
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        envvar="TWC_CONFIG_FILE",
        show_envvar=False,
        exists=True,
        dir_okay=False,
        callback=config_callback,
        hidden=True,
        help="Use config.",
    ),
    locate: bool = typer.Option(
        False, "--locate", help="Open file directory in file manager."
    ),
):
    """Print config file path."""
    print(config)
    if locate:
        typer.launch(str(config), locate=True)


# ------------------------------------------------------------- #
# $ twc config dump                                             #
# ------------------------------------------------------------- #


class DumpFormat(str, Enum):
    """Configuration file format variants."""

    TOML = "toml"
    YAML = "yaml"
    JSON = "json"


@config.command("dump")
def config_dump(
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        envvar="TWC_CONFIG_FILE",
        show_envvar=False,
        exists=True,
        dir_okay=False,
        callback=config_callback,
        hidden=True,
        help="Use config.",
    ),
    profile: Optional[str] = profile_option,
    output_format: DumpFormat = typer.Option(
        DumpFormat.TOML.value,
        "--output",
        "-o",
        case_sensitive=False,
        help="Output format.",
    ),
    full_dump: bool = typer.Option(
        False, "--full", help="Dump full configuration."
    ),
):
    """Dump configuration."""
    if full_dump:
        config_dict = load_config(config)
    else:
        config_dict = load_config(config)[profile]

    encoders = {
        "toml": toml.dumps,
        "yaml": yaml.dump,
        "json": json.dumps,
    }

    print_colored(
        encoders[output_format.value](config_dict), lang=output_format.value
    )


# ------------------------------------------------------------- #
# $ twc config set                                              #
# ------------------------------------------------------------- #


def validate_params_callback(value: list) -> list:
    """Validate CLI configuration parameters."""
    for param in value:
        if not re.match(r"^[a-z0-9_]+=[a-z0-9-_.]+$", param, re.I):
            raise typer.BadParameter(
                f"'{param}': Parameter must be a KEY=VALUE pair."
            )
    return value


@config.command("set")
def config_set(
    params: List[str] = typer.Argument(
        ...,
        callback=validate_params_callback,
        help="List of KEY=VALUE parameters.",
    ),
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Set config parameters."""
    config_dict = load_config(config)
    params_dict = {}
    for param in params:
        key, value = param.split("=")
        if value.isdigit():
            value = int(value)
        params_dict[key] = value
    config_dict[profile] = merge_dicts(config_dict[profile], params_dict)
    write_to_file(config_dict, config)


# ------------------------------------------------------------- #
# $ twc config unset                                            #
# ------------------------------------------------------------- #


@config.command("unset")
def config_unset(
    params: List[str] = typer.Argument(..., help="List of parameters."),
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Unset config parameters."""
    config_dict = load_config(config)
    for param in params:
        try:
            del config_dict[profile][param]
        except KeyError:
            sys.exit(f"Error: No such key: '{param}'")
    write_to_file(config_dict, config)


# ------------------------------------------------------------- #
# $ twc config edit                                             #
# ------------------------------------------------------------- #


@config.command("edit")
def config_edit(
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        envvar="TWC_CONFIG_FILE",
        show_envvar=False,
        exists=True,
        dir_okay=False,
        callback=config_callback,
        hidden=True,
        help="Use config.",
    ),
):
    """Open configuration file in default editor."""
    typer.launch(str(config))


# ------------------------------------------------------------- #
# $ twc config profiles                                         #
# ------------------------------------------------------------- #


@config.command("profiles")
def config_profile(
    # pylint: disable=redefined-outer-name
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        envvar="TWC_CONFIG_FILE",
        show_envvar=False,
        exists=True,
        dir_okay=False,
        callback=config_callback,
        hidden=True,
        help="Use config.",
    ),
):
    """Display configuration profiles."""
    config_dict = load_config(config)
    for key in config_dict.keys():
        print(key)



"""Manage databases."""

import re
import sys
import textwrap
from datetime import date, datetime
from typing import Optional, List
from pathlib import Path
from ipaddress import IPv4Address, IPv4Network

import typer
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.api import ServiceRegion, MySQLAuthPlugin, BackupInterval
from twc.apiwrap import create_client
from twc.vars import REGION_ZONE_MAP
from twc.utils import merge_dicts
from .common import (
    verbose_option,
    config_option,
    profile_option,
    filter_option,
    yes_option,
    output_format_option,
    load_from_config_callback,
    zone_option,
)


database = TyperAlias(help=__doc__)
database_backup = TyperAlias(help="Manage database backups.")
database.add_typer(database_backup, name="backup")
database_user = TyperAlias(help="Manage database users.")
database.add_typer(database_user, name="user")
database_instance = TyperAlias(
    help="Manage instances in cluster (databases/topics/etc)."
)
database.add_typer(database_instance, name="instance", aliases=["db"])


# ------------------------------------------------------------- #
# $ twc database list                                           #
# ------------------------------------------------------------- #


def print_databases(response: Response, filters: Optional[str]):
    """Print table with databases list."""
    # pylint: disable=invalid-name
    dbs = response.json()["dbs"]
    if filters:
        dbs = fmt.filter_list(dbs, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "STATUS",
            "TYPE",
            "IPV4",
            "INTERNAL IP",
        ]
    )
    for db in dbs:
        table.row(
            [
                db["id"],
                db["name"],
                db["status"],
                db["type"],
                db["ip"],
                db["local_ip"],
            ]
        )
    table.print()


@database.command("list", "ls")
def database_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    limit: int = typer.Option(500, help="Items to display."),
    filters: Optional[str] = filter_option,
):
    """List databases."""
    client = create_client(config, profile)
    response = client.get_databases(limit=limit)
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_databases,
    )


# ------------------------------------------------------------- #
# $ twc database get                                            #
# ------------------------------------------------------------- #


def print_database(response: Response):
    """Print table with database info."""
    # pylint: disable=invalid-name
    db = response.json()["db"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "STATUS",
            "TYPE",
            "IPV4",
            "INTERNAL IP",
        ]
    )
    table.row(
        [
            db["id"],
            db["name"],
            db["status"],
            db["type"],
            db["ip"],
            db["local_ip"],
        ]
    )
    table.print()


@database.command("get")
def database_get(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    status: Optional[bool] = typer.Option(
        False,
        "--status",
        help="Display status and exit with 0 if status is 'started'.",
    ),
):
    """Get database info."""
    client = create_client(config, profile)
    response = client.get_database(db_id)
    if status:
        state = response.json()["db"]["status"]
        if state == "started":
            print(state)
            raise typer.Exit()
        sys.exit(state)
    fmt.printer(response, output_format=output_format, func=print_database)


# ------------------------------------------------------------- #
# $ twc database list-presets                                   #
# ------------------------------------------------------------- #


def print_dbs_presets(response: Response, filters: Optional[str]):
    """Print table with database presets list."""
    presets = response.json()["databases_presets"]
    if filters:
        presets = fmt.filter_list(presets, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "REGION",
            "PRICE",
            "CPU",
            "RAM",
            "DISK",
            "TYPE",
        ]
    )
    for preset in presets:
        table.row(
            [
                preset["id"],
                preset["location"],
                preset["price"],
                preset["cpu"],
                str(round(preset["ram"] / 1024)) + "G",
                str(round(preset["disk"] / 1024)) + "G",
                preset["type"],
            ]
        )
    table.print()


@database.command("list-presets", "lp")
def database_list_presets(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
    region: Optional[ServiceRegion] = typer.Option(
        None,
        case_sensitive=False,
        help="Use region (location).",
    ),
):
    """List database configuration presets."""
    if region:
        filters = f"{filters},location:{region}"
    client = create_client(config, profile)
    response = client.get_database_presets()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_dbs_presets,
    )


# ------------------------------------------------------------- #
# $ twc database list-types                                     #
# ------------------------------------------------------------- #


@database.command("list-types", "lt")
def database_list_types(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List database configuration presets."""
    client = create_client(config, profile)
    response = client.get_database_types().json()
    table = fmt.Table()
    table.header(["TYPE", "DATABASE", "VERSION"])
    for service in response["types"]:
        table.row(
            [
                service["type"],
                service["name"],
                service["version"],
            ]
        )
    table.print()


# ------------------------------------------------------------- #
# $ twc database create                                         #
# ------------------------------------------------------------- #


def set_params(params: list) -> dict:
    """Return dict with database config_parameters."""
    parameters = {}
    for param in params:
        if re.match(r"^([a-z_]+)=([0-9a-zA-Z]+)$", param):
            parameter, value = param.split("=")
            if value.isdigit():
                value = int(value)
            parameters[parameter] = value
        else:
            raise typer.BadParameter(
                f"'{param}': Parameter can contain only digits,"
                " latin letters and underscore."
            )
    return parameters


def dbms_parameters_callback(value: str) -> List[str]:
    if value:
        return value.split(",")
    return []


@database.command("create")
def database_create(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    availability_zone: Optional[str] = zone_option,
    preset_id: int = typer.Option(..., help="Database configuration preset."),
    dbms: str = typer.Option(
        ...,
        "--type",
        help="Database management system. See TYPE in `twc database list-types`.",
    ),
    hash_type: Optional[MySQLAuthPlugin] = typer.Option(
        MySQLAuthPlugin.CACHING_SHA2.value,
        case_sensitive=False,
        help="Authentication plugin for MySQL.",
    ),
    name: str = typer.Option(..., help="Database cluster display name."),
    params: Optional[List[str]] = typer.Option(
        None,
        "--param",
        metavar="PARAM=VALUE",
        help="Database config parameters, can be multiple.",
    ),
    login: Optional[str] = typer.Option(
        # DEPRECATED
        None,
        hidden=True,
        help="Database user login.",
    ),
    password: Optional[str] = typer.Option(
        # DEPRECATED
        None,
        hidden=True,
    ),
    user_login: Optional[str] = typer.Option(None, help="User login."),
    user_password: Optional[str] = typer.Option(None, help="User password."),
    user_host: Optional[str] = typer.Option(
        "%", help="User host for MySQL, Postgres"
    ),
    user_privileges: Optional[str] = typer.Option(
        None,
        help="Comma-separated list of user privileges.",
        callback=dbms_parameters_callback,
    ),
    user_desc: Optional[str] = typer.Option(None, help="Comment for user."),
    db_name: Optional[str] = typer.Option(None, help="Database name."),
    db_desc: Optional[str] = typer.Option(None, help="Database comment."),
    network_id: Optional[str] = typer.Option(None, help="Private network ID."),
    private_ip: Optional[str] = typer.Option(
        None, help="Private IPv4 address."
    ),
    public_ip: Optional[str] = typer.Option(
        None, help="Public IPv4 address. New address by default."
    ),
    no_public_ip: Optional[bool] = typer.Option(
        False, "--no-public-ip", help="Do not add public IPv4 address."
    ),
    project_id: Optional[int] = typer.Option(
        None,
        envvar="TWC_PROJECT",
        show_envvar=False,
        callback=load_from_config_callback,
        help="Add database cluster to specific project.",
    ),
    enable_backups: Optional[bool] = typer.Option(
        False,
        "--enable-backups",
        help="Enable atomatic backups of database cluster.",
    ),
    backup_keep: Optional[int] = typer.Option(
        1,
        show_default=True,
        help="Number of backups to keep.",
    ),
    backup_start_date: Optional[datetime] = typer.Option(
        date.today().strftime("%Y-%m-%d"),
        formats=["%Y-%m-%d"],
        show_default=False,
        help="Start date of the first backup creation [default: today].",
    ),
    backup_interval: Optional[BackupInterval] = typer.Option(
        BackupInterval.DAY.value,
        "--backup-interval",
        help="Backup interval.",
    ),
    backup_day_of_week: Optional[int] = typer.Option(
        1,
        min=1,
        max=7,
        help="The day of the week on which backups will be created."
        " NOTE: This option works only with interval 'week'."
        " First day of week is monday.",
    ),
):
    """Create managed database cluster."""
    client = create_client(config, profile)

    payload = {
        "dbms": dbms,
        "preset_id": preset_id,
        "name": name,
        "hash_type": hash_type,
        "config_parameters": {},
        "network": {},
    }

    if enable_backups:
        backup_start_date = backup_start_date.strftime("%Y-%m-%dT%H:%M:%SZ")
        payload["auto_backups"] = {
            "copy_count": backup_keep,
            "creation_started_at": backup_start_date,
            "interval": backup_interval,
            "day_of_week": backup_day_of_week,
        }

    if network_id:
        payload["network"]["id"] = network_id
        if private_ip:
            net = IPv4Network(
                client.get_vpc(network_id).json()["vpc"]["subnet_v4"]
            )
            if IPv4Address(private_ip) >= net.network_address + 4:
                payload["network"]["ip"] = private_ip
            else:
                # First 3 addresses is reserved by Timeweb Cloud for gateway and future use.
                sys.exit(
                    f"Error: Private address '{private_ip}' is not allowed. "
                    "IP must be at least the fourth in order in the network."
                )
    if public_ip:
        try:
            _ = IPv4Address(public_ip)
            payload["network"]["floating_ip"] = public_ip
        except ValueError:
            sys.exit(f"Error: '{public_ip}' is not valid IPv4 address.")
    else:
        # Get new public IPv4 address.
        if no_public_ip is False:
            zone = None
            if preset_id and not availability_zone:
                for preset in client.get_database_presets().json()[
                    "databases_presets"
                ]:
                    if preset["id"] == preset_id:
                        zone = REGION_ZONE_MAP[preset["location"]]
            if availability_zone:
                zone = availability_zone
            ip = client.create_floating_ip(availability_zone=zone).json()[
                "ip"
            ]["ip"]
            payload["network"]["floating_ip"] = ip

    if login:
        print(
            "--login is deprecated use --user-login instead", file=sys.stderr
        )
        user_login = login

    if password:
        print(
            "--password is deprecated use --user-password instead",
            file=sys.stderr,
        )
        user_password = password

    if user_password and not user_login:
        sys.exit("Error: --user-login required if --user-password is set.")

    if user_login:
        if not user_password:
            user_password = typer.prompt(
                "Database user password",
                hide_input=True,
                confirmation_prompt=True,
            )
        payload["admin"] = {
            "login": user_login,
            "password": user_password,
            "host": user_host,
            "privileges": user_privileges,
            "description": user_desc or "",
        }

    if db_name:
        payload["instance"] = {
            "name": db_name,
            "description": db_desc or "",
        }

    if params:
        payload["config_parameters"] = set_params(params)

    if project_id:
        payload["project_id"] = project_id

    response = client.create_database2(**payload)

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["db"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc database set                                            #
# ------------------------------------------------------------- #


@database.command("set")
def database_set(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    preset_id: Optional[int] = typer.Option(
        None, help="Database configuration preset."
    ),
    external_ip: bool = typer.Option(
        True, help="Enable external IPv4 address."
    ),
    name: Optional[str] = typer.Option(
        None, help="Database instance display name."
    ),
    params: Optional[List[str]] = typer.Option(
        None,
        "--param",
        metavar="PARAM=VALUE",
        help="Database parameters, can be multiple.",
    ),
    password: Optional[str] = typer.Option(
        None, help="Database user password"
    ),
    prompt_password: Optional[bool] = typer.Option(
        False, help="Set database user password interactively."
    ),
):
    """Set database properties and parameters."""
    client = create_client(config, profile)
    old_state = client.get_database(db_id).json()["db"]
    new_params = {}
    if prompt_password:
        password = typer.prompt(
            "Database user password",
            hide_input=True,
            confirmation_prompt=True,
        )
    payload = {
        "preset_id": preset_id,
        "name": name,
        "password": password,
        "config_parameters": {},
        "external_ip": external_ip,
    }
    if params:
        new_params = set_params(params)
    if new_params:
        payload["config_parameters"] = merge_dicts(
            old_state["config_parameters"], new_params
        )
    response = client.update_database(db_id, **payload)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["db"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc database remove                                         #
# ------------------------------------------------------------- #


@database.command("remove", "rm")
def database_remove(
    db_ids: List[int] = typer.Argument(..., metavar="DB_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove database."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for db_id in db_ids:
        response = client.delete_database(db_id)
        if response.status_code == 200:
            del_hash = response.json()["database_delete"]["hash"]
            del_code = typer.prompt("Please enter confirmation code", type=int)
            response = client.delete_database(
                db_id, delete_hash=del_hash, code=del_code
            )
        if response.status_code == 204:
            print(db_id)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc database backup list                                    #
# ------------------------------------------------------------- #


def print_db_backups(response: Response):
    """Print table with DB backup list."""
    backups = response.json()["backups"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "DISK",
            "CREATED",
            "STATUS",
        ]
    )
    for bak in backups:
        table.row(
            [
                bak["id"],
                bak["name"],
                bak["created_at"],
                bak["status"],
            ]
        )
    table.print()


@database_backup.command("list", "ls")
def database_backup_list(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List backups."""
    client = create_client(config, profile)
    response = client.get_database_backups(db_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=print_db_backups,
    )


# ------------------------------------------------------------- #
# $ twc database backup create                                  #
# ------------------------------------------------------------- #


@database_backup.command("create")
def database_backup_create(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Backup database."""
    client = create_client(config, profile)
    response = client.create_database_backup(db_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["backup"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc database backup remove                                  #
# ------------------------------------------------------------- #


@database_backup.command("remove", "rm")
def database_backup_remove(
    db_id: int,
    backup_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove backup."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    response = client.delete_database_backup(db_id, backup_id)
    if response.status_code == 204:
        print(backup_id)
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc database backup restore                                 #
# ------------------------------------------------------------- #


@database_backup.command("restore")
def database_backup_restore(
    db_id: int,
    backup_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Restore backup."""
    if not yes:
        typer.confirm("Data on target disk will lost. Continue?", abort=True)
    client = create_client(config, profile)
    response = client.restore_database_backup(db_id, backup_id)
    if response.status_code == 204:
        print(backup_id)
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc database backup schedule                                #
# ------------------------------------------------------------- #


def print_autobackup_settings(response: Response):
    """Print backup settings info."""
    table = fmt.Table()
    settings = response.json()["auto_backups_settings"]
    translated_keys = {
        "copy_count": "Keep copies",
        "creation_start_at": "Backup start date",
        "is_enabled": "Enabled",
        "interval": "Interval",
        "day_of_week": "Day of week",
    }
    for key in settings.keys():
        table.row([translated_keys[key], ":", settings[key]])
    table.print()


@database_backup.command("schedule")
def database_backup_schedule(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    status: bool = typer.Option(
        False,
        "--status",
        help="Display automatic backups status.",
    ),
    enable: Optional[bool] = typer.Option(
        None,
        "--enable/--disable",
        show_default=False,
        help="Enable or disable automatic backups.",
    ),
    keep: int = typer.Option(
        1,
        show_default=True,
        help="Number of backups to keep.",
    ),
    start_date: datetime = typer.Option(
        date.today().strftime("%Y-%m-%d"),
        formats=["%Y-%m-%d"],
        show_default=False,
        help="Start date of the first backup creation [default: today].",
    ),
    interval: BackupInterval = typer.Option(
        BackupInterval.DAY.value,
        "--interval",
        help="Backup interval.",
    ),
    day_of_week: Optional[int] = typer.Option(
        1,
        min=1,
        max=7,
        help="The day of the week on which backups will be created."
        " NOTE: This option works only with interval 'week'."
        " First day of week is monday.",
    ),
):
    """Manage database cluster automatic backup settings."""
    client = create_client(config, profile)

    if status:
        response = client.get_database_autobackup_settings(db_id)
        fmt.printer(
            response,
            output_format=output_format,
            func=print_autobackup_settings,
        )
        if response.json()["auto_backups_settings"]["is_enabled"]:
            sys.exit(0)
        else:
            sys.exit(1)

    start_date = start_date.strftime("%Y-%m-%dT%H:%M:%SZ")

    response = client.update_database_autobackup_settings(
        db_id,
        is_enabled=enable,
        copy_count=keep,
        creation_start_at=start_date,
        interval=interval,
        day_of_week=day_of_week,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=print_autobackup_settings,
    )


# ------------------------------------------------------------- #
# $ twc database user list                                      #
# ------------------------------------------------------------- #


def _print_database_users(response: Response):
    users = response.json()["admins"]
    table = fmt.Table()
    table.header(["ID", "LOGIN", "HOST", "CREATED"])
    for user in users:
        table.row(
            [
                user["id"],
                user["login"],
                user["host"],
                user["created_at"],
            ]
        )
    table.print()


@database_user.command("list", "ls")
def database_user_list(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List database users."""
    client = create_client(config, profile)
    response = client.get_database_users(db_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=_print_database_users,
    )


# ------------------------------------------------------------- #
# $ twc database user get                                       #
# ------------------------------------------------------------- #


def _print_database_user(response: Response):
    user = response.json()["admin"]
    out = textwrap.dedent(
        f"""
        Login:       {user['login']}
        Host:        {user['host']}
        Created:     {user['created_at']}
        Description: {user['description']}
        """
    ).strip()
    print(out)
    print()
    table = fmt.Table()
    table.header(["INSTANCE_ID", "PRIVILEGES"])
    for instance in user["instances"]:
        table.row(
            [
                instance["instance_id"],
                ", ".join(instance["privileges"]),
            ]
        )
    table.print()


@database_user.command("get")
def database_user_get(
    db_id: int,
    user_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get database user."""
    client = create_client(config, profile)
    response = client.get_database_user(db_id, user_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=_print_database_user,
    )


# ------------------------------------------------------------- #
# $ twc database user create                                    #
# ------------------------------------------------------------- #


@database_user.command("create")
def database_user_create(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    login: str = typer.Option(..., help="User login."),
    password: str = typer.Option(
        ...,
        prompt=True,
        hide_input=True,
        confirmation_prompt=True,
        help="User password.",
    ),
    host: Optional[str] = typer.Option(
        "%", help="User host for MySQL, Postgres"
    ),
    instance_id: Optional[int] = typer.Option(
        None,
        help="The specific instance ID to which the privileges will be "
        "applied. If not specified, the privileges will be applied to "
        "all available instances.",
    ),
    privileges: Optional[str] = typer.Option(
        [],
        help="Comma-separated list of user privileges.",
        callback=dbms_parameters_callback,
    ),
    desc: Optional[str] = typer.Option(None, help="Comment for user."),
):
    """Create database users."""
    client = create_client(config, profile)
    response = client.create_database_user(
        db_id=db_id,
        login=login,
        password=password,
        privileges=privileges,
        host=host,
        instance_id=instance_id,
        description=desc,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["admin"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc database user remove                                    #
# ------------------------------------------------------------- #


@database_user.command("remove", "rm")
def database_user_remove(
    db_id: int,
    user_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Delete database user."""
    client = create_client(config, profile)
    response = client.get_database_user(db_id, user_id)
    if response.status_code == 204:
        print(user_id)
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc database instance list                                  #
# ------------------------------------------------------------- #


def _print_database_instances(response: Response):
    instances = response.json()["instances"]
    table = fmt.Table()
    table.header(["ID", "NAME", "CREATED", "DESCRIPTION"])
    for i in instances:
        table.row(
            [
                i["id"],
                i["name"],
                i["created_at"],
                i["description"],
            ]
        )
    table.print()


@database_instance.command("list", "ls")
def database_instance_list(
    db_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List databases in database cluster."""
    client = create_client(config, profile)
    response = client.get_database_instances(db_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=_print_database_instances,
    )


# ------------------------------------------------------------- #
# $ twc database instance create                                #
# ------------------------------------------------------------- #


@database_instance.command("create")
def database_instance_create(
    db_id: int,
    name: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    desc: Optional[str] = typer.Option(None, help="Comment for database."),
):
    """Create database in database cluster."""
    client = create_client(config, profile)
    response = client.create_database_instance(
        db_id, name=name, description=desc
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["instance"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc database instance remove                                #
# ------------------------------------------------------------- #


@database_instance.command("remove", "rm")
def database_instance_remove(
    db_id: int,
    instance_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Delete database from cluster."""
    client = create_client(config, profile)
    response = client.get_database_user(db_id, instance_id)
    if response.status_code == 204:
        print(instance_id)
    else:
        sys.exit(fmt.printer(response))



"""API client exceptions."""

from typing import List, Union
from uuid import UUID

from requests import HTTPError, PreparedRequest, Response


class ErrResponse:
    """API error response schema."""

    def __init__(
        self,
        status_code: int = None,
        error_code: str = None,
        message: Union[str, List[str]] = None,
        response_id: UUID = None,
        **kwargs,
    ):
        self.status_code = status_code
        self.error_code = error_code
        self.message = message
        self.response_id = response_id
        self.kwargs = kwargs


class TimewebCloudException(HTTPError):
    """Base exception for TimewebCloud."""

    def __init__(
        self,
        request: PreparedRequest = None,
        response: Response = None,
        error_code: str = None,
        status_code: str = None,
        message: Union[str, List[str]] = None,
        response_id: UUID = None,
    ):
        if message is not None:
            if isinstance(message, list):
                description = "; ".join(message)
            else:
                description = message
        else:
            description = response.reason
        super().__init__(description)
        self.request = request
        self.response = response
        self.error_code = error_code
        self.status_code = status_code
        self.message = description
        self.response_id = response_id


class UnauthorizedError(TimewebCloudException):
    """User is unauthorized. Mostly user does not have API access token."""


class MalformedResponseError(TimewebCloudException):
    """API respond non JSON response body, or with malformed JSON schema."""


class UnexpectedResponseError(TimewebCloudException):
    """API respond unexpected response. E.g. 502 Bad Gateway, etc."""


class BadRequestError(TimewebCloudException):
    """400 Bad Request."""


class ForbiddenError(TimewebCloudException):
    """403 Forbidden."""


class NotFoundError(TimewebCloudException):
    """404 Not Found."""


class ConflictError(TimewebCloudException):
    """409 Conflict."""


class LockedError(TimewebCloudException):
    """423 Locked."""


class TooManyRequestsError(TimewebCloudException):
    """429 Too Many Requests."""


class InternalServerError(TimewebCloudException):
    """500 Internal Server Error."""


class NetworkError(Exception):
    """Raises on requests.exceptions.ConnectionError."""



"""Manage SSH-keys."""

import sys
from typing import Optional, List
from pathlib import Path

import typer
from click import UsageError
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.apiwrap import create_client
from .common import (
    verbose_option,
    config_option,
    profile_option,
    yes_option,
    output_format_option,
)


ssh_key = TyperAlias(help=__doc__)


# ------------------------------------------------------------- #
# $ twc ssh-key list                                            #
# ------------------------------------------------------------- #


def print_ssh_keys(response: Response):
    """Print table with SSH-keys list."""
    table = fmt.Table()
    table.header(["ID", "NAME", "DEFAULT", "SERVERS"])
    keys = response.json()["ssh_keys"]
    for key in keys:
        table.row(
            [
                key["id"],
                key["name"],
                key["is_default"],
                ", ".join([str(k["id"]) for k in key["used_by"]]),
            ]
        )
    table.print()


@ssh_key.command("list", "ls")
def ssh_key_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List SSH-keys."""
    client = create_client(config, profile)
    response = client.get_ssh_keys()
    fmt.printer(
        response,
        output_format=output_format,
        func=print_ssh_keys,
    )


# ------------------------------------------------------------- #
# $ twc ssh-key get                                             #
# ------------------------------------------------------------- #


def print_ssh_key(response: Response):
    """Print table with SSH-key info."""
    table = fmt.Table()
    table.header(["ID", "NAME", "DEFAULT", "SERVERS"])
    key = response.json()["ssh_key"]
    table.row(
        [
            key["id"],
            key["name"],
            key["is_default"],
            ", ".join([str(k["id"]) for k in key["used_by"]]),
        ]
    )
    table.print()


@ssh_key.command("get")
def ssh_key_get(
    ssh_key_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get SSH-key by ID."""
    client = create_client(config, profile)
    response = client.get_ssh_key(ssh_key_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=print_ssh_key,
    )


# ------------------------------------------------------------- #
# $ twc ssh-key new                                             #
# ------------------------------------------------------------- #


@ssh_key.command("new")
def ssh_key_new(
    public_key_file: Path = typer.Argument(
        ...,
        metavar="FILE",
        exists=True,
        dir_okay=False,
    ),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(None, help="SSH-key display name."),
    default: Optional[bool] = typer.Option(
        False,
        "--default",
        help="Set as default key for new Cloud Servers.",
    ),
):
    """Upload new SSH-key."""
    if not name:
        name = Path(public_key_file).name

    try:
        with open(public_key_file, "r", encoding="utf-8") as pubkey:
            body = pubkey.read().strip()
    except (OSError, IOError, FileNotFoundError) as error:
        sys.exit(f"Error: {error}")

    client = create_client(config, profile)
    response = client.add_new_ssh_key(
        name=name,
        is_default=default,
        body=body,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["ssh_key"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc ssh-key edit                                            #
# ------------------------------------------------------------- #


@ssh_key.command("edit", "set", "update", "upd")
def ssh_key_edit(
    ssh_key_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(None, help="SSH-key display name."),
    public_key_file: Optional[Path] = typer.Option(
        None,
        metavar="FILE",
        exists=True,
        dir_okay=False,
        help="Public key file.",
    ),
    default: Optional[bool] = typer.Option(
        False,
        "--default",
        help="Set as default key for new Cloud Servers.",
    ),
):
    """Edit SSH-key and they properties."""
    client = create_client(config, profile)
    body = None

    if public_key_file:
        try:
            with open(public_key_file, "r", encoding="utf-8") as pubkey:
                body = pubkey.read().strip()
        except (OSError, IOError, FileNotFoundError) as error:
            sys.exit(f"Error: {error}")

    response = client.update_ssh_key(
        ssh_key_id,
        name=name,
        body=body,
        is_default=default,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["ssh_key"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc ssh-key add                                             #
# ------------------------------------------------------------- #


@ssh_key.command("add", "copy")
def ssh_key_add(
    server_id: int = typer.Argument(...),
    ssh_keys_ids: List[int] = typer.Argument(..., metavar="SSH_KEY_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Copy SSH-keys to Cloud Server."""
    client = create_client(config, profile)
    response = client.add_ssh_key_to_server(server_id, ssh_keys_ids)
    if response.status_code == 204:
        print(server_id)
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc ssh-key remove                                          #
# ------------------------------------------------------------- #


@ssh_key.command("remove", "rm")
def ssh_key_remove(
    ssh_keys_ids: List[int] = typer.Argument(..., metavar="SSH_KEY_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    server_id: Optional[int] = typer.Option(
        None,
        "--from-server",
        metavar="SERVER_ID",
        help="Remove SSH-key from server instead of remove key itself.",
    ),
    yes: Optional[bool] = yes_option,
):
    """Remove SSH-keys."""
    if not yes:
        typer.confirm(
            "If you do not specify '--from-server' option "
            "SSH-key will be deleted\nfrom all servers where it was added "
            "and also deleted itself.\nContinue?",
            abort=True,
        )
    client = create_client(config, profile)
    if server_id:
        if len(ssh_keys_ids) >= 2:
            raise UsageError("Cannot remove multiple keys from server.")
        response = client.delete_ssh_key_from_server(
            server_id, ssh_keys_ids[0]
        )
        if response.status_code == 204:
            print(ssh_keys_ids[0])
            raise typer.Exit()
        sys.exit(fmt.printer(response))

    for ssh_key_id in ssh_keys_ids:
        response = client.delete_ssh_key(ssh_key_id)
        if response.status_code == 204:
            print(ssh_key_id)
        else:
            sys.exit(fmt.printer(response))



"""Manage Kubernetes clusters."""

import re
import sys
from typing import Optional, List
from pathlib import Path

import typer
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.api import TimewebCloud
from twc.apiwrap import create_client
from .common import (
    verbose_option,
    config_option,
    profile_option,
    filter_option,
    yes_option,
    output_format_option,
    load_from_config_callback,
)


cluster = TyperAlias(help=__doc__)
cluster_group = TyperAlias(help="Manage worker node groups.")
cluster_node = TyperAlias(help="Manage worker nodes.")
cluster.add_typer(cluster_group, name="group", aliases=["groups"])
cluster.add_typer(cluster_node, name="node", aliases=["nodes"])


# ------------------------------------------------------------- #
# $ twc cluster list                                            #
# ------------------------------------------------------------- #


def print_clusters(response: Response, filters: Optional[str]):
    """Print table with K8s clusters list."""
    clusters = response.json()["clusters"]
    if filters:
        clusters = fmt.filter_list(clusters, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "STATUS",
            "VERSION",
        ]
    )
    for cl in clusters:
        table.row(
            [
                cl["id"],
                cl["name"],
                cl["status"],
                cl["k8s_version"],
            ]
        )
    table.print()


@cluster.command("list", "ls")
def cluster_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List Kubernetes clusters."""
    client = create_client(config, profile)
    response = client.get_k8s_clusters()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_clusters,
    )


# ------------------------------------------------------------- #
# $ twc cluster remove                                          #
# ------------------------------------------------------------- #


@cluster.command("remove", "rm")
def cluster_remove(
    cluster_ids: List[int] = typer.Argument(..., metavar="CLUSTER_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove Kubernetes cluster."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for cluster_id in cluster_ids:
        response = client.delete_k8s_cluster(cluster_id)
        if response.status_code == 200:
            del_hash = response.json()["cluster_delete"]["hash"]
            del_code = typer.prompt("Please enter confirmation code", type=int)
            response = client.delete_k8s_cluster(
                cluster_id, delete_hash=del_hash, code=del_code
            )
        if response.status_code == 204:
            print(cluster_id)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc cluster create                                          #
# ------------------------------------------------------------- #


@cluster.command("create")
def cluster_create(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: str = typer.Option(..., help="Cluster's display name."),
    desc: str = typer.Option(None, help="Cluster description."),
    k8s_version: Optional[str] = typer.Option(
        None,
        help="Kubernetes version."
        " See 'twc k8s list-k8s-versions'. Latest by deafult.",
    ),
    master_preset_id: Optional[int] = typer.Option(
        None,
        help="Master node configuration preset. Minimal by default.",
    ),
    network_driver: Optional[str] = typer.Option(
        "canal",
        help="Network driver.",
    ),
    ingress: bool = typer.Option(True, help="Enable Nginx ingress."),
    worker_groups: List[str] = typer.Option(
        None,
        "--add-worker-group",
        metavar="NAME,PRESET_ID,NODE_COUNT",
        help="Add workers node group.",
    ),
    project_id: Optional[int] = typer.Option(
        None,
        envvar="TWC_PROJECT",
        show_envvar=False,
        callback=load_from_config_callback,
        help="Add cluster to specific project.",
    ),
):
    """Create Kubernetes cluster."""
    client = create_client(config, profile)

    if worker_groups:
        actual_worker_groups = []
        pattern = (
            "^((name=)?[a-z0-9_-]+),((preset=)?[0-9]+),((count=)?[0-9]+)$"
        )
        for wg in worker_groups:
            if not re.match(pattern, wg, re.I):
                sys.exit(
                    "Error: Wrong worker group pattern.\n"
                    "Pattern must include comma sepatated sections:\n"
                    " - name (group name, only a-zA-Z0-9-_)\n"
                    " - preset (nodes preset_id, number)\n"
                    " - count (number of workers)\n"
                    "Correct pattern: "
                    "'name=my_group1,preset=399,count=2' or 'my_group1,399,2'"
                )
            _name, _preset, _count = wg.split(",")
            worker_group = {
                "name": _name.replace("name=", ""),
                "preset_id": int(_preset.replace("preset=", "")),
                "node_count": int(_count.replace("count=", "")),
            }
            actual_worker_groups.append(worker_group)
        worker_groups = actual_worker_groups

    # Select latest K8s version
    if not k8s_version:
        versions = client.get_k8s_versions().json()["k8s_versions"]
        versions.sort(
            key=lambda s: [int(u.replace("v", "")) for u in s.split(".")],
            reverse=True,
        )
        k8s_version = versions[0]  # apply latest version

    # Select minimal master node preset excluding promo preset.
    # Promo preset has 'limit=1'.
    if not master_preset_id:
        presets = client.get_k8s_presets().json()["k8s_presets"]
        master_presets = [
            p for p in presets if p["type"] == "master" and p["limit"] is None
        ]
        master_presets_ram = [p["ram"] for p in master_presets]
        for preset in master_presets:
            if preset["ram"] == min(master_presets_ram):
                master_preset_id = preset["id"]

    if project_id:
        if not project_id in [
            prj["id"] for prj in client.get_projects().json()["projects"]
        ]:
            sys.exit(f"Wrong project ID: Project '{project_id}' not found.")

    response = client.create_k8s_cluster(
        name=name,
        description=desc,
        k8s_version=k8s_version,
        network_driver=network_driver,
        ingress=ingress,
        worker_groups=worker_groups,
        preset_id=master_preset_id,
    )

    # Add created cluster to project if set
    if project_id:
        client.add_cluster_to_project(
            response.json()["cluster"]["id"],
            project_id,
        )

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["cluster"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc cluster kubeconfig                                      #
# ------------------------------------------------------------- #


@cluster.command("kubeconfig", "kubecfg", "cfg")
def cluster_kubeconfig(
    cluster_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    save: Optional[Path] = typer.Option(
        None,
        help="Path to file. NOTE: Existing file will be overwitten.",
    ),
):
    """Download KubeConfig."""
    client = create_client(config, profile)
    file_content = client.get_k8s_cluster_kubeconfig(cluster_id).text
    if save:
        with open(save, "w", encoding="utf-8") as kubeconfig:
            kubeconfig.write(file_content)
    else:
        fmt.print_colored(file_content, lang="yaml")


# ------------------------------------------------------------- #
# $ twc cluster list-k8s-versions                               #
# ------------------------------------------------------------- #


@cluster.command("list-k8s-versions", "lv")
def cluster_list_versions(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List available Kubernetes versions."""
    client = create_client(config, profile)
    response = client.get_k8s_versions()
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: [
            print(v) for v in response.json()["k8s_versions"]
        ],
    )


# ------------------------------------------------------------- #
# $ twc cluster list-network-drivers                            #
# ------------------------------------------------------------- #


@cluster.command("list-network-drivers")
def cluster_list_network_drivers(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List available Kubernetes network drivers."""
    client = create_client(config, profile)
    response = client.get_k8s_network_drivers()
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: [
            print(v) for v in response.json()["network_drivers"]
        ],
    )


# ------------------------------------------------------------- #
# $ twc cluster list-presets                                    #
# ------------------------------------------------------------- #


def print_presets(response: Response, filters: Optional[str] = None):
    """Print table with K8s nodes presets."""
    presets = response.json()["k8s_presets"]
    if filters:
        presets = fmt.filter_list(presets, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "PRICE",
            "CPU",
            "RAM",
            "DISK",
            "TYPE",
        ]
    )
    for preset in presets:
        table.row(
            [
                preset["id"],
                preset["price"],
                preset["cpu"],
                str(round(preset["ram"] / 1024)) + "G",
                str(preset["disk"]) + "G",
                preset["type"],
            ]
        )
    table.print()


@cluster.command("list-presets", "lp")
def cluster_list_presets(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List nodes configuration presets."""
    client = create_client(config, profile)
    response = client.get_k8s_presets()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_presets,
    )


# ------------------------------------------------------------- #
# $ twc cluster show                                           #
# ------------------------------------------------------------- #


def print_cluster_info(response: Response):
    info = response.json()["resources"]
    table = fmt.Table()
    table.header(
        [
            "RESOURCE",
            "REQUESTED",
            "ALLOCATABLE",
            "CAPACITY",
            "USED",
        ]
    )
    resources = list(info.keys())
    resources.remove("nodes")
    for r in resources:
        table.row(
            [
                r,
                info[r]["requested"],
                info[r]["allocatable"],
                info[r]["capacity"],
                info[r]["used"],
            ]
        )
    table.print()


@cluster.command("show")
def cluster_show(
    cluster_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Show cluster resource usage."""
    client = create_client(config, profile)
    response = client.get_k8s_cluster_resources(cluster_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=print_cluster_info,
    )


# ------------------------------------------------------------- #
# $ twc cluster group list                                      #
# ------------------------------------------------------------- #


def print_node_groups(response: Response):
    groups = response.json()["node_groups"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "CREATED",
            "PRESET",
            "NODES",
        ]
    )
    for group in groups:
        table.row(
            [
                group["id"],
                group["name"],
                group["created_at"],
                group["preset_id"],
                group["node_count"],
            ]
        )
    table.print()


@cluster_group.command("list", "ls")
def cluster_group_list(
    cluster_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List cluster node groups."""
    client = create_client(config, profile)
    response = client.get_k8s_node_groups(cluster_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=print_node_groups,
    )


# ------------------------------------------------------------- #
# $ twc cluster group add                                       #
# ------------------------------------------------------------- #


@cluster_group.command("add")
def cluster_group_add(
    cluster_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: str = typer.Option(..., help="Node group display name."),
    preset_id: int = typer.Option(..., help="Nodes configuration preset."),
    nodes: int = typer.Option(1, help="Number of nodes in group."),
):
    """Add new node group."""
    client = create_client(config, profile)
    response = client.create_k8s_node_group(
        cluster_id,
        name=name,
        preset_id=preset_id,
        node_count=nodes,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["node_group"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc cluster group remove                                    #
# ------------------------------------------------------------- #


def get_cluster_id_by_group(client: TimewebCloud, group_id: int) -> int:
    clusters = client.get_k8s_clusters().json()["clusters"]
    for _cluster in clusters:
        groups = client.get_k8s_node_groups(_cluster["id"]).json()[
            "node_groups"
        ]
        for group in groups:
            if group_id == group["id"]:
                return _cluster["id"]
    sys.exit(f"Error: Node group '{group_id}' not found.")


@cluster_group.command("remove", "rm")
def cluster_group_remove(
    group_ids: List[int] = typer.Argument(..., metavar="GROUP_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove node group."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for group_id in group_ids:
        cluster_id = get_cluster_id_by_group(client, group_id)
        response = client.delete_k8s_node_group(cluster_id, group_id)
        if response.status_code == 204:
            print(group_id)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc cluster group scale-up                                  #
# ------------------------------------------------------------- #


@cluster_group.command("scale-up")
def cluster_group_scaleup(
    group_id: int,
    count: Optional[int] = typer.Argument(1, help="Nodes count."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Add worker nodes to group."""
    client = create_client(config, profile)
    cluster_id = get_cluster_id_by_group(client, group_id)
    response = client.add_k8s_nodes_to_group(cluster_id, group_id, count)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(
            f"Nodes in group {group_id}:", response.json()["meta"]["total"]
        ),
    )


# ------------------------------------------------------------- #
# $ twc cluster group scale-down                                #
# ------------------------------------------------------------- #


@cluster_group.command("scale-down")
def cluster_group_scaledown(
    group_id: int,
    count: Optional[int] = typer.Argument(1, help="Nodes count."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Remove worker nodes from group."""
    client = create_client(config, profile)
    cluster_id = get_cluster_id_by_group(client, group_id)
    response = client.delete_k8s_nodes_from_group(cluster_id, group_id, count)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(
            f"Nodes in group {group_id}:", response.json()["meta"]["total"]
        ),
    )


# ------------------------------------------------------------- #
# $ twc cluster node list                                       #
# ------------------------------------------------------------- #


def print_nodes(response: Response):
    nodes = response.json()["nodes"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "GROUP",
            "TYPE",
            "STATUS",
            "CREATED",
            "CPU/RAM/DISK",
        ]
    )
    for node in nodes:
        table.row(
            [
                node["id"],
                node["group_id"],
                node["type"],
                node["status"],
                node["created_at"],
                str(node["cpu"])
                + "/"
                + str(round(node["ram"] / 1024))
                + "G/"
                + str(node["disk"])
                + "G",
            ]
        )
    table.print()


@cluster_node.command("list", "ls")
def cluster_node_list(
    cluster_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List cluster node groups."""
    client = create_client(config, profile)
    response = client.get_k8s_nodes(cluster_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=print_nodes,
    )


# ------------------------------------------------------------- #
# $ twc cluster node remove                                     #
# ------------------------------------------------------------- #


@cluster_node.command("remove", "rm")
def cluster_node_remove(
    cluster_id: int,
    node_ids: List[int] = typer.Argument(..., metavar="NODE_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove nodes from cluster."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for node_id in node_ids:
        response = client.delete_k8s_node(cluster_id, node_id)
        if response.status_code == 204:
            print(node_id)
        else:
            sys.exit(fmt.printer(response))



"""Common functions for commands."""

import re
import sys
from enum import Enum
from typing import Optional, Any
from pathlib import Path, PurePath
from logging import basicConfig, debug, DEBUG

import toml
import typer
from typer.core import TyperOption
from click import UsageError

from twc.__version__ import __version__
from twc.api.types import ServiceRegion, ServiceAvailabilityZone


class OutputFormat(str, Enum):
    """Data output formats. See `output_format_option`."""

    DEFAULT = "default"
    RAW = "raw"
    JSON = "json"
    YAML = "yaml"


def get_root_ctx(ctx: typer.Context) -> typer.Context:
    """Return the context of the root command. This function allows you
    to retrieve the `typer.Context` object for the root command using the
    current context.

    Each command has its own context, which stores data about the command
    itself, all options and arguments entered by the user, and other service
    information.

    Each `typer.Context` object has a `parent` field that contains a
    reference to the parent context. Thus it is possible to collect all
    available contexts.

    Specifically, this function loops through the objects in the `parent`
    field until it reaches the root context, whose `parent` is None and
    returns it.

    See *_callback functions below for examples.
    """
    context = ctx
    while True:
        debug(f"cli_parser: cmd={context.command}, params={context.params}")
        if context.parent is None:
            break
        context = context.parent
    return context


def default_config_file() -> Path:
    """Return default configuration file path."""
    filenames = [
        ".twcrc",
        ".twcrc.toml",
    ]
    for filename in filenames:
        filepath = Path(PurePath(Path.home()).joinpath(filename))
        if filepath.exists():
            debug(f"Configuration file found: {filepath}")
            return filepath
    return Path(PurePath(Path.home()).joinpath(filenames[0]))


def load_config(filepath: Optional[Path] = default_config_file()) -> dict:
    """Load configuration from TOML config file."""
    try:
        with open(filepath, "r", encoding="utf-8") as file:
            return toml.load(file)
    except FileNotFoundError:
        sys.exit(
            f"Configuration file {filepath} not found. Try run 'twc config'"
        )
    except OSError as error:
        sys.exit(f"Error: Cannot read configuration file {filepath}: {error}")
    except toml.TomlDecodeError as error:
        sys.exit(f"Error: Check your TOML syntax in file {filepath}: {error}")


def load_from_config_callback(
    param: TyperOption, value: Any, ctx: typer.Context
) -> Any:
    """Return value from config file. NOTE: This callback doesn't work with
    custom enumerations such as OutputFormat and may others. Typer will return
    None instead of actual value returning by this function.
    """
    if value is None:
        try:
            config = config_callback(ctx.params["config"], ctx)
            profile = profile_callback(ctx.params["profile"], ctx)
            debug(
                f"Set '{param.name}' value from file:"
                f" config={config}, profile={profile}"
            )
            value = load_config(config)[profile][param.name]
        except KeyError as err:
            debug(f"Not found key {err} in: {config}:{profile}")
            return None
    debug(f"Return value: {value}")
    if param.name == "output_format":
        value = value.lower()
    if param.name == "region":
        value = value.lower()
        regions = [v.value for v in ServiceRegion]
        if value not in regions:
            sys.exit(f"Error: Location not in {regions}")
    if param.name == "zone":
        value = value.lower()
        zones = [z.value for z in ServiceAvailabilityZone]
        if value not in zones:
            sys.exit(f"Error: Availability zone not in {zones}")
    return value


def version_callback(value: bool):
    """Print version and exit."""
    if value:
        print(f"v{__version__}")
        raise typer.Exit()


def verbose_callback(value: bool) -> bool:
    """Typer callback to enable verbose mode. Set logger settings."""
    if value:
        basicConfig(
            force=True,
            level=DEBUG,
            format="%(asctime)s:%(levelname)s:%(name)s:%(module)s:%(funcName)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
        )
    return value


def config_callback(value: Path, ctx: typer.Context) -> Path:
    """Lookup for '--config' option value and save it into `ctx`."""
    if not value:
        root_value = get_root_ctx(ctx).params["config"]
        debug(f"cli parser: root lvl: config={root_value}")
        value = root_value or default_config_file()
    return value


def profile_callback(value: str, ctx: typer.Context) -> str:
    """Lookup for '--profile' option value and save it into `ctx`."""
    if not value:
        root_value = get_root_ctx(ctx).params["profile"]
        debug(f"cli parser: root lvl: profile={root_value}")
        value = root_value or "default"
    return value


def filter_callback(value: str) -> str:
    """Validate filters format."""
    if value:
        if not re.match(
            r"^(([a-z0-9._-]+:[a-z0-9._\-/+%]+),?)+$", value, re.I
        ):
            raise UsageError(
                "Invalid filter format. Filter must contain comma separated "
                + "KEY:VALUE pairs. Allowed characters: [a-zA-Z0-9_.-/+%]"
            )
        # Make aliases
        value = value.replace("region:", "location:")
        value = value.replace("proto:", "protocol:")
        value = value.replace("rule_id:", "id:")
    return value


# ------------------------------------------------------------- #
# Very common CLI options.                                      #
# ------------------------------------------------------------- #


# version: Optional[bool] = version_option,

version_option = typer.Option(
    None,
    "--version",
    callback=version_callback,
    is_eager=True,
    help="Show version and exit.",
)


# verbose: Optional[bool] = verbose_option,

verbose_option = typer.Option(
    False,
    "--verbose",
    "-v",
    envvar="TWC_DEBUG",
    show_envvar=False,
    callback=verbose_callback,
    help="Enable verbose mode.",
    rich_help_panel="Global options",
)


# Do not use following 'config' and 'profile' options in __main__.py.
# They callbacks doesn't works properly at root level command e.g. twc -c ...


# config: Optional[Path] = config_option,

config_option = typer.Option(
    None,
    "--config",
    "-c",
    envvar="TWC_CONFIG_FILE",
    show_envvar=False,
    show_default=False,
    exists=True,
    dir_okay=False,
    callback=config_callback,
    help="Use config.",
    rich_help_panel="Global options",
)


# profile: Optional[str] = profile_option,

profile_option = typer.Option(
    None,
    "--profile",
    "-p",
    metavar="NAME",
    envvar="TWC_PROFILE",
    show_envvar=False,
    show_default=False,
    callback=profile_callback,
    help="Use profile.",
    rich_help_panel="Global options",
)


# output_format: Optional[str] = output_format_option,

output_format_option = typer.Option(
    None,
    "--output",
    "-o",
    metavar="FORMAT",
    envvar="TWC_OUTPUT_FORMAT",
    show_default=False,
    show_envvar=False,
    callback=load_from_config_callback,
    help=f"Output format, one of: [{'|'.join([k.value for k in OutputFormat])}].",
)


# filters: Optional[str] = filter_option,

filter_option = typer.Option(
    None,
    "--filter",
    "-f",
    metavar="KEY:VALUE",
    callback=filter_callback,
    help="Filter output.",
)


# yes: Optional[bool] = yes_option,

yes_option = typer.Option(
    False, "--yes", "-y", help="Confirm the action without prompting."
)


# region: Optional[str] = region_option,

region_option = typer.Option(
    "ru-1",
    metavar="REGION",
    envvar="TWC_REGION",
    show_envvar=False,
    callback=load_from_config_callback,
    help="Region (location).",
)

# availability_zone: Optional[str] = zone_option,

zone_option = typer.Option(
    None,
    metavar="ZONE",
    envvar="TWC_AVAILABILITY_ZONE",
    show_envvar=False,
    callback=load_from_config_callback,
    help="Availability zone.",
)



"""API Client wrapper for CLI."""

import os
import sys
import textwrap
from pathlib import Path
from logging import debug, warning

from .api import TimewebCloud
from .api import exceptions as exc
from .commands.common import load_config


def request_handler(func):
    """Error handler decorator for requests. Wrap API exceptions
    and exit with human-readable error message.
    """

    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except exc.NetworkError as err:
            sys.exit(f"Error: {err}")
        except exc.UnauthorizedError as err:
            sys.exit(
                textwrap.dedent(
                    f"""
                    Error: {err}
                    Please check your API access token. Try run 'twc config'
                    """
                ).strip()
            )
        except exc.MalformedResponseError as err:
            sys.exit(
                textwrap.dedent(
                    f"""
                    Error: API returned malformed response: {err}
                    Try run command with '--verbose' option for details.
                    """
                ).strip()
            )
        except exc.UnexpectedResponseError as err:
            sys.exit(
                textwrap.dedent(
                    f"""
                    Error: API returned unexpected response: {err}
                    Try run command with '--verbose' option for details.
                    """
                ).strip()
            )
        except exc.TimewebCloudException as err:
            sys.exit(
                textwrap.dedent(
                    f"""
                    Error ocurred.
                    Status code: {err.status_code}
                    Error code: {err.error_code}
                    Message: {err.message}
                    Response ID: {err.response_id}
                    """
                ).strip()
            )

    return wrapper


def create_client(config: Path, profile: str, **kwargs) -> TimewebCloud:
    """API client wrapper. Read configuration file and return
    `TimewebCloud` object with decorator.
    """
    token = os.getenv("TWC_TOKEN")
    log_settings = os.getenv("TWC_LOG")
    api_endpoint = os.getenv("TWC_ENDPOINT")

    if api_endpoint:
        warning("Using API URL from environment: %s", api_endpoint)
        kwargs["api_base_url"] = api_endpoint

    if log_settings:
        for param in log_settings.split(","):
            if param.lower() == "debug":
                pass  # FUTURE: set logging.DEBUG

    if token:
        debug("Config: use API token from environment")
    else:
        try:
            debug(f"Config: file={config}, profile={profile}")
            token = load_config(config)[profile]["token"]
        except KeyError:
            sys.exit(f"Error: Profile '{profile}' not found in {config}")

    return TimewebCloud(token, request_decorator=request_handler, **kwargs)



"""Manage load balancers."""

import re
import sys
from enum import Enum
from typing import Optional, List
from pathlib import Path
from logging import debug
from ipaddress import IPv4Address, IPv4Network

import typer
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.apiwrap import create_client
from twc.api import TimewebCloud
from twc.vars import REGION_ZONE_MAP
from twc.api.types import LoadBalancerAlgo, LoadBalancerProto, ServiceRegion
from .common import (
    verbose_option,
    config_option,
    profile_option,
    filter_option,
    yes_option,
    region_option,
    zone_option,
    output_format_option,
    load_from_config_callback,
)


balancer = TyperAlias(help=__doc__)
balancer_backend = TyperAlias(help="Manage load balancer backend servers.")
balancer_rule = TyperAlias(help="Manage load balancer rules.")
balancer.add_typer(balancer_backend, name="backend", aliases=["backends"])
balancer.add_typer(balancer_rule, name="rule", aliases=["rules"])


# ------------------------------------------------------------- #
# $ twc balancer list                                           #
# ------------------------------------------------------------- #


def print_balancers(response: Response, filters: Optional[str]):
    """Print table with balancers list."""
    balancers = response.json()["balancers"]
    if filters:
        balancers = fmt.filter_list(balancers, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "STATUS",
            "INTERNAL IP",
            "EXTERNAL IP",
        ]
    )
    for lb in balancers:
        table.row(
            [
                lb["id"],
                lb["name"],
                lb["status"],
                lb["local_ip"],
                lb["ip"],
            ]
        )
    table.print()


@balancer.command("list", "ls")
def balancer_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List load balancers."""
    client = create_client(config, profile)
    response = client.get_load_balancers()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_balancers,
    )


# ------------------------------------------------------------- #
# $ twc balancer get                                            #
# ------------------------------------------------------------- #


def print_balancer(response: Response):
    """Print table with load balancer info."""
    lb = response.json()["balancer"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "STATUS",
            "INTERNAL IP",
            "EXTERNAL IP",
        ]
    )
    table.row(
        [
            lb["id"],
            lb["name"],
            lb["status"],
            lb["local_ip"],
            lb["ip"],
        ]
    )
    table.print()


@balancer.command("get")
def balancer_get(
    balancer_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    status: Optional[bool] = typer.Option(
        False,
        "--status",
        help="Display status and exit with 0 if status is 'started'.",
    ),
):
    """Get load balancer info."""
    client = create_client(config, profile)
    response = client.get_load_balancer(balancer_id)
    if status:
        state = response.json()["balancer"]["status"]
        if state == "started":
            print(state)
            raise typer.Exit()
        sys.exit(state)
    fmt.printer(response, output_format=output_format, func=print_balancer)


# ------------------------------------------------------------- #
# $ twc balancer create                                         #
# ------------------------------------------------------------- #


class CertType(str, Enum):
    """..."""

    CUSTOM = "custom"
    LETS_ENCRYPT = "lets_encrypt"


@balancer.command("create")
def balancer_create(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: str = typer.Option(..., help="Load balancer display name."),
    desc: Optional[str] = typer.Option(
        None, help="Load balancer description."
    ),
    preset_id: Optional[int] = typer.Option(
        None, help="Load balancer preset ID."
    ),
    replicas: int = typer.Option(
        1,
        min=1,
        max=2,
        help="Load balancer replica count. Ignored if --preset-id set.",
    ),
    algo: LoadBalancerAlgo = typer.Option(
        LoadBalancerAlgo.ROUND_ROBIN.value,
        help="Balancer algorythm.",
    ),
    port: int = typer.Option(80, help="Load balancer listen port."),
    path: str = typer.Option("/", help="URL path."),
    proto: LoadBalancerProto = typer.Option(
        LoadBalancerProto.HTTP.value, help="Health check protocol."
    ),
    inter: int = typer.Option(10, help="Health checks interval in seconds."),
    timeout: int = typer.Option(5, help="Health check timeout in seconds."),
    rise: int = typer.Option(
        3,
        help="Number of successful health checks to consider backend as operational.",
    ),
    fall: int = typer.Option(
        2,
        help="Number of unsuccessfull health checks to consider backend as dead.",
    ),
    sticky: bool = typer.Option(False, help="Stick on client IP."),
    proxy_protocol: bool = typer.Option(False),
    force_https: bool = typer.Option(False),
    backend_keepalive: bool = typer.Option(False),
    max_connections: Optional[int] = typer.Option(
        None, help="Backend server's maximum number of concurrent connections."
    ),
    connect_timeout: Optional[int] = typer.Option(
        None,
        help="Maximum time to wait for a connection attempt to a backend server to succeed.",
    ),
    client_timeout: Optional[int] = typer.Option(
        None, help="Maximum inactivity time on the client side."
    ),
    server_timeout: Optional[int] = typer.Option(
        None, help="Maximum time for pending data staying into output buffer."
    ),
    http_timeout: Optional[int] = typer.Option(
        None, help="Maximum allowed time to wait for a complete HTTP request."
    ),
    project_id: Optional[int] = typer.Option(
        None,
        envvar="TWC_PROJECT",
        show_envvar=False,
        callback=load_from_config_callback,
        help="Add load balancer to specific project.",
    ),
    network: Optional[str] = typer.Option(
        None, hidden=True, help="Private network ID."
    ),
    network_id: Optional[str] = typer.Option(None, help="Private network ID."),
    private_ip: Optional[str] = typer.Option(
        None, help="Private IPv4 address."
    ),
    public_ip: Optional[str] = typer.Option(
        None, help="Public IPv4 address. New address by default."
    ),
    no_public_ip: Optional[bool] = typer.Option(
        False, "--no-public-ip", help="Do not add public IPv4 address."
    ),
    region: Optional[str] = region_option,
    availability_zone: Optional[str] = zone_option,
    cert_type: Optional[CertType] = typer.Option(
        None,
        help="SSL certificate type. Falls to 'custom' "
        "if --cert-data and --cert-key set.",
    ),
    cert_domain: Optional[str] = typer.Option(
        None,
        help="Domain name for which the certificate was issued. "
        "Note: domain name A-record will set to load balancer's public IP.",
    ),
    cert_data: Optional[typer.FileText] = typer.Option(
        None, help="Fullchain certificate file."
    ),
    cert_key: Optional[typer.FileText] = typer.Option(
        None, help="Certificate key file."
    ),
):
    """Create load balancer."""
    client = create_client(config, profile)

    payload = {
        "name": name,
        "comment": desc,
        "preset_id": preset_id,
        "algo": algo,
        "proto": proto,
        "port": port,
        "path": path,
        "inter": inter,
        "timeout": timeout,
        "fall": fall,
        "rise": rise,
        "sticky": sticky,
        "proxy_protocol": proxy_protocol,
        "force_https": force_https,
        "backend_keepalive": backend_keepalive,
        "network": {},
        "project_id": project_id,
        "certificates": {},
        "max_connections": max_connections,
        "connect_timeout": connect_timeout,
        "client_timeout": client_timeout,
        "server_timeout": server_timeout,
        "http_timeout": http_timeout,
    }

    if cert_type == CertType.CUSTOM:
        if not cert_data or not cert_key:
            sys.exit(
                "Error: --cert-data and --cert-key is required if --cert-type is 'custom'"
            )
    elif cert_type == CertType.LETS_ENCRYPT:
        if cert_data or cert_key:
            sys.exit(
                "Error: --cert-data and --cert-key is not allowed with --cert-type 'lets_encrypt'"
            )
    if not cert_type:
        if cert_data and not cert_key:
            sys.exit("Error: --cert-key is required.")
        if cert_key and not cert_data:
            sys.exit("Error: --cert-data is required.")
        if cert_data and cert_key:
            cert_type = CertType.CUSTOM
    if cert_type and not cert_domain:
        sys.exit(
            "Error: --cert-domain is required if --cert-type and/or --cert-data, --cert-key is set."
        )
    if cert_type:
        payload["certificates"] = {
            "type": cert_type.value,
            "fqdn": cert_domain,
            **({"cert_data": cert_data.read()} if cert_data else {}),
            **({"key_data": cert_key.read()} if cert_key else {}),
        }

    if not preset_id:
        for preset in client.get_load_balancer_presets().json()[
            "balancers_presets"
        ]:
            if (
                preset["replica_count"] == replicas
                and preset["location"] == region
            ):
                preset_id = preset["id"]
        if not preset_id:
            sys.exit(f"Error: Cannot set {replicas} load balancer replicas.")

    if network:
        print(
            "Option --network is deprecated and will be removed soon, "
            "use --network-id instead",
            file=sys.stderr,
        )
        network_id = network

    if network_id:
        payload["network"]["id"] = network_id
        if private_ip:
            net = IPv4Network(
                client.get_vpc(network_id).json()["vpc"]["subnet_v4"]
            )
            if IPv4Address(private_ip) >= net.network_address + 4:
                payload["network"]["ip"] = private_ip
            else:
                # First 3 addresses is reserved by Timeweb Cloud for gateway and future use.
                sys.exit(
                    f"Error: Private address '{private_ip}' is not allowed. "
                    "IP must be at least the fourth in order in the network."
                )
    if public_ip:
        try:
            _ = IPv4Address(public_ip)
            payload["network"]["floating_ip"] = public_ip
        except ValueError:
            sys.exit(f"Error: '{public_ip}' is not valid IPv4 address.")
    else:
        # Get new public IPv4 address.
        if no_public_ip is False:
            zone = None
            if not preset_id and not availability_zone and not region:
                sys.exit(
                    "Error: Unable to get IPv4 address, at least one of "
                    "[--preset-id, --region, --availability-zone] "
                    "must be set to determine correct location."
                )
            if region:
                zone = REGION_ZONE_MAP[region]
            if preset_id and not availability_zone:
                for preset in client.get_database_presets().json()[
                    "databases_presets"
                ]:
                    if preset["id"] == preset_id:
                        zone = REGION_ZONE_MAP[preset["location"]]
            if availability_zone:
                zone = availability_zone
            ip = client.create_floating_ip(availability_zone=zone).json()[
                "ip"
            ]["ip"]
            payload["network"]["floating_ip"] = ip

    # Create LB
    response = client.create_load_balancer(**payload)

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["balancer"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc balancer set                                            #
# ------------------------------------------------------------- #


@balancer.command("set")
def balancer_set(
    balancer_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(
        None, help="Load balancer display name."
    ),
    replicas: Optional[int] = typer.Option(
        None,
        min=1,
        max=2,
        help="Load balancer replica count.",
    ),
    algo: LoadBalancerAlgo = typer.Option(None, help="Balancer algorythm."),
    port: Optional[int] = typer.Option(
        None, help="Load balancer listen port."
    ),
    path: Optional[str] = typer.Option(None, help="URL path."),
    proto: Optional[LoadBalancerProto] = typer.Option(
        None, help="Health check protocol."
    ),
    inter: Optional[int] = typer.Option(
        None, help="Health checks interval in seconds."
    ),
    timeout: Optional[int] = typer.Option(
        None, help="Health check timeout in seconds."
    ),
    rise: Optional[int] = typer.Option(
        None,
        help="Number of successful health checks to consider backend as operational.",
    ),
    fall: Optional[int] = typer.Option(
        None,
        help="Number of unsuccessfull health checks to consider backend as dead.",
    ),
    sticky: Optional[bool] = typer.Option(None, help="Stick on client IP."),
    proxy_protocol: Optional[bool] = typer.Option(None),
    force_https: Optional[bool] = typer.Option(None),
    backend_keepalive: Optional[bool] = typer.Option(None),
    max_connections: Optional[int] = typer.Option(
        None, help="Backend server's maximum number of concurrent connections."
    ),
    connect_timeout: Optional[int] = typer.Option(
        None,
        help="Maximum time to wait for a connection attempt to a backend server to succeed.",
    ),
    client_timeout: Optional[int] = typer.Option(
        None, help="Maximum inactivity time on the client side."
    ),
    server_timeout: Optional[int] = typer.Option(
        None, help="Maximum time for pending data staying into output buffer."
    ),
    http_timeout: Optional[int] = typer.Option(
        None, help="Maximum allowed time to wait for a complete HTTP request."
    ),
):
    """Change load balancer parameters."""
    client = create_client(config, profile)

    preset_id = None
    if replicas is not None:
        for preset in client.get_load_balancer_presets().json()[
            "balancers_presets"
        ]:
            if preset["replica_count"] == replicas:
                preset_id = preset["id"]
        if not preset_id:
            sys.exit(f"Error: Cannot set {replicas} load balancer replicas.")

    response = client.update_load_balancer(
        balancer_id=balancer_id,
        name=name,
        preset_id=preset_id,
        algo=algo,
        proto=proto,
        port=port,
        path=path,
        inter=inter,
        timeout=timeout,
        fall=fall,
        rise=rise,
        sticky=sticky,
        proxy_protocol=proxy_protocol,
        force_https=force_https,
        backend_keepalive=backend_keepalive,
        max_connections=max_connections,
        connect_timeout=connect_timeout,
        client_timeout=client_timeout,
        server_timeout=server_timeout,
        http_timeout=http_timeout,
    )

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["balancer"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc balancer remove                                         #
# ------------------------------------------------------------- #


@balancer.command("remove", "rm")
def balancer_remove(
    balancer_ids: List[int] = typer.Argument(..., metavar="BALANCER_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove load balancer."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for balancer_id in balancer_ids:
        response = client.delete_load_balancer(balancer_id)
        if response.status_code == 200:
            del_hash = response.json()["balancer_delete"]["hash"]
            del_code = typer.prompt("Please enter confirmation code", type=int)
            response = client.delete_load_balancer(
                balancer_id, delete_hash=del_hash, code=del_code
            )
        if response.status_code == 204:
            print(balancer_id)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc balancer backed list                                    #
# ------------------------------------------------------------- #


@balancer_backend.command("list", "ls")
def blancer_backend_list(
    balancer_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List load balancer backends."""
    client = create_client(config, profile)
    response = client.get_load_balancer_ips(balancer_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print("\n".join(response.json()["ips"])),
    )


# ------------------------------------------------------------- #
# $ twc balancer list-presets                                     #
# ------------------------------------------------------------- #


def _print_presets(response: Response, filters: Optional[str] = None):
    presets = response.json()["balancers_presets"]
    if filters:
        presets = fmt.filter_list(presets, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "REGION",
            "PRICE",
            "REPLICAS",
            "BANDW",
            "RPS",
        ]
    )
    for preset in presets:
        table.row(
            [
                preset["id"],
                preset["location"],
                preset["price"],
                preset["replica_count"],
                preset["bandwidth"],
                preset["request_per_second"],
            ]
        )
    table.print()


@balancer.command("list-presets", "lp")
def balancer_list_presets(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
    region: Optional[ServiceRegion] = typer.Option(
        None, help="Use region (location)."
    ),
):
    """List configuration presets."""
    if region:
        filters = f"{filters},location:{region}"
    client = create_client(config, profile)
    response = client.get_load_balancer_presets()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=_print_presets,
    )


# ------------------------------------------------------------- #
# $ twc balancer backend add                                    #
# ------------------------------------------------------------- #


@balancer_backend.command("add")
def blancer_backend_add(
    balancer_id: int = typer.Argument(...),
    backends: List[str] = typer.Argument(..., metavar="BACKEND_IP..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Add new backend servers to balancer."""
    client = create_client(config, profile)
    response = client.add_ips_to_load_balancer(balancer_id, backends)
    if response.status_code == 204:
        print("\n".join(backends))
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc balancer backed remove                                  #
# ------------------------------------------------------------- #


@balancer_backend.command("remove", "rm")
def blancer_backend_remove(
    balancer_id: int = typer.Argument(...),
    backends: List[str] = typer.Argument(..., metavar="BACKEND..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove load balancer backends."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    response = client.delete_ips_from_load_balancer(balancer_id, backends)
    if response.status_code == 204:
        print("\n".join(backends))
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc balancer rule list                                      #
# ------------------------------------------------------------- #


def print_balancer_rules(response: Response):
    """Print balancers list."""
    rules = response.json()["rules"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "FRONTEND PORT",
            "FRONTEND PROTO",
            "BACKEND PORT",
            "BACKEND PROTO",
        ]
    )
    for rule in rules:
        table.row(
            [
                rule["id"],
                rule["balancer_port"],
                rule["balancer_proto"],
                rule["server_port"],
                rule["server_proto"],
            ]
        )
    table.print()


@balancer_rule.command("list", "ls")
def balancer_rule_list(
    balancer_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List load balancer rules."""
    client = create_client(config, profile)
    response = client.get_load_balancer_rules(balancer_id)
    fmt.printer(
        response,
        output_format=output_format,
        func=print_balancer_rules,
    )


# ------------------------------------------------------------- #
# $ twc balancer rule add                                       #
# ------------------------------------------------------------- #


def validate_balancer_port_proto_callback(value):
    """Typer callback for PORT/PROTO validation."""
    if value:
        if not re.match(r"^\d+\/(http|http2|https|tcp)$", value, re.I):
            sys.exit(
                f"Error: Malformed argument: '{value}': "
                "Correct patterns: '443/HTTPS', '2000/TCP', etc.\n"
                "Available protocols: "
                f"{', '.join([lb.value for lb in LoadBalancerProto])}"
            )
        port, proto = value.split("/")
        port = int(port)
        return port, proto
    return value


@balancer_rule.command("add")
def balancer_rule_add(
    balancer_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    frontend: str = typer.Option(
        ...,
        callback=validate_balancer_port_proto_callback,
        metavar="PORT/PROTO",
        help="Frontend port and protocol.",
    ),
    backend: str = typer.Option(
        ...,
        callback=validate_balancer_port_proto_callback,
        metavar="PORT/PROTO",
        help="Backend port and protocol.",
    ),
):
    """Add load balancer rule."""
    client = create_client(config, profile)

    f_port, f_proto = frontend
    b_port, b_proto = backend

    response = client.create_load_balancer_rule(
        balancer_id,
        balancer_port=f_port,
        balancer_proto=f_proto,
        server_port=b_port,
        server_proto=b_proto,
    )

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["rule"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc balancer rule update                                    #
# ------------------------------------------------------------- #


def get_balancer_id_by_rule(client: TimewebCloud, rule_id: int) -> int:
    """Return load balancer ID by balancer rule ID."""
    balancers = client.get_load_balancers().json()["balancers"]
    for lb in balancers:
        rules = client.get_load_balancer_rules(lb["id"]).json()["rules"]
        for rule in rules:
            debug(f"{rule['id']} from {lb['id']} compare with {rule_id}")
            if rule["id"] == rule_id:
                return lb["id"]
    sys.exit(f"Error: Rule '{rule_id}' not found.")


@balancer_rule.command("update", "upd")
def balancer_rule_update(
    rule_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    frontend: Optional[str] = typer.Option(
        None,
        callback=validate_balancer_port_proto_callback,
        metavar="PORT/PROTO",
        help="Frontend port and protocol.",
    ),
    backend: Optional[str] = typer.Option(
        None,
        callback=validate_balancer_port_proto_callback,
        metavar="PORT/PROTO",
        help="Backend port and protocol.",
    ),
):
    """Update load balancer rule."""
    client = create_client(config, profile)

    f_port = f_proto = None
    b_port = b_proto = None

    balancer_id = get_balancer_id_by_rule(client, rule_id)
    old_rules = client.get_load_balancer_rules(balancer_id).json()["rules"]
    for old_rule in old_rules:
        if old_rule["id"] == rule_id:
            f_port = old_rule["balancer_port"]
            f_proto = old_rule["balancer_proto"]
            b_port = old_rule["server_port"]
            b_proto = old_rule["server_proto"]

    if frontend:
        f_port, f_proto = frontend
    if backend:
        b_port, b_proto = backend

    response = client.update_load_balancer_rule(
        balancer_id,
        rule_id,
        balancer_port=f_port,
        balancer_proto=f_proto,
        server_port=b_port,
        server_proto=b_proto,
    )

    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["rule"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc balancer rule remove                                    #
# ------------------------------------------------------------- #


@balancer_rule.command("remove", "rm")
def balancer_rule_remove(
    rule_ids: List[int] = typer.Argument(..., metavar="RULE_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove load balancer rule."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for rule_id in rule_ids:
        balancer_id = get_balancer_id_by_rule(client, rule_id)
        response = client.delete_load_balancer_rule(balancer_id, rule_id)
        if response.status_code == 204:
            print(rule_id)
        else:
            sys.exit(fmt.printer(response))



"""Manage virtual networks."""

import sys
from typing import Optional, List
from pathlib import Path
from ipaddress import (
    IPv4Network,
    AddressValueError,
    NetmaskValueError,
)

import typer
from requests import Response

from twc import fmt
from twc.api import ServiceRegion
from twc.typerx import TyperAlias
from twc.apiwrap import create_client
from twc.vars import REGIONS_WITH_LAN, ZONES_WITH_LAN
from .common import (
    verbose_option,
    config_option,
    profile_option,
    yes_option,
    output_format_option,
    filter_option,
    region_option,
    zone_option,
)


vpc = TyperAlias(help=__doc__)
vpc_port = TyperAlias(help="Manage network ports.")
vpc.add_typer(vpc_port, name="port", aliases=["ports"])

ALLOWED_SUBNETS = [
    IPv4Network("10.0.0.0/8"),
    IPv4Network("192.168.0.0/16"),
    IPv4Network("172.16.0.0/12"),
]


# ------------------------------------------------------------- #
# $ twc vpc list                                                #
# ------------------------------------------------------------- #


def print_networks(response: Response, filters: Optional[str] = None):
    nets = response.json()["vpcs"]
    if filters:
        nets = fmt.filter_list(nets, filters)
    table = fmt.Table()
    table.header(["ID", "REGION", "ZONE", "SUBNET"])
    for net in nets:
        table.row(
            [
                net["id"],
                net["location"],
                net["availability_zone"],
                net["subnet_v4"],
            ]
        )
    table.print()


@vpc.command("list", "ls")
def vpc_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List networks."""
    client = create_client(config, profile)
    response = client.get_vpcs()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_networks,
    )


# ------------------------------------------------------------- #
# $ twc vpc create                                              #
# ------------------------------------------------------------- #


def validate_network(value):
    if value:
        try:
            network = IPv4Network(value)
        except (NetmaskValueError, AddressValueError, ValueError) as err:
            sys.exit(f"Error: Invalid CIDR: {err}")

        is_valid = False
        for subnet in ALLOWED_SUBNETS:
            if network.subnet_of(subnet):
                is_valid = True

        if is_valid is False:
            sys.exit(
                f"Error: Network {value} is not subnet of: "
                f"{[n.with_prefixlen for n in ALLOWED_SUBNETS]}"
            )
    return value


@vpc.command("create")
def vpc_create(
    subnet: str = typer.Argument(
        ...,
        metavar="NETWORK_SUBNET",
        callback=validate_network,
        help="IPv4 network CIDR.",
    ),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: str = typer.Option(None, help="Network display name."),
    desc: Optional[str] = typer.Option(None, help="Description."),
    region: Optional[str] = region_option,
    availability_zone: Optional[str] = zone_option,
):
    """Create network."""
    client = create_client(config, profile)
    if region not in REGIONS_WITH_LAN:
        sys.exit(
            f"Error: Cannot create network in location '{region}'. "
            f"Available regions are: {REGIONS_WITH_LAN}"
        )
    usable_zones = set(ZONES_WITH_LAN).intersection(
        set(ServiceRegion.get_zones(region))
    )
    if availability_zone is not None and availability_zone not in usable_zones:
        sys.exit(
            f"Error: Cannot create network in region '{region}' with "
            f"availability zone '{availability_zone}'. "
            f"Usable zones are: {list(usable_zones)}"
        )
    if not name:
        name = subnet
    response = client.create_vpc(
        name,
        description=desc,
        subnet=subnet,
        location=region,
        availability_zone=availability_zone,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["vpc"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc vpc remove                                              #
# ------------------------------------------------------------- #


@vpc.command("remove", "rm")
def vpc_remove(
    vpc_ids: List[str] = typer.Argument(..., metavar="VPC_ID..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    yes: Optional[bool] = yes_option,
):
    """Remove network."""
    if not yes:
        typer.confirm("This action cannot be undone. Continue?", abort=True)
    client = create_client(config, profile)
    for vpc_id in vpc_ids:
        response = client.delete_vpc(vpc_id)
        if response.status_code == 204:
            print(vpc_id)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc vpc set                                                 #
# ------------------------------------------------------------- #


@vpc.command("set")
def vpc_set(
    vpc_id: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    name: Optional[str] = typer.Option(None, help="Network display name."),
    desc: Optional[str] = typer.Option(None, help="Description."),
):
    """Set network properties."""
    client = create_client(config, profile)
    response = client.update_vpc(vpc_id, name=name, description=desc)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["vpc"]["id"]),
    )


# ------------------------------------------------------------- #
# $ twc vpc port list                                           #
# ------------------------------------------------------------- #


def print_ports(response: Response, filters: Optional[str] = None):
    ports = response.json()["vpc_ports"]
    if filters:
        ports = fmt.filter_list(ports, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAT",
            "IP",
        ]
    )
    for port in ports:
        table.row(
            [
                port["id"],
                port["nat_mode"],
                port["ipv4"],
            ]
        )
    table.print()


@vpc_port.command("list", "ls")
def vpc_ports_list(
    vpc_id: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List network ports."""
    client = create_client(config, profile)
    response = client.get_vpc_ports(vpc_id)
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_ports,
    )


# ------------------------------------------------------------- #
# $ twc vpc show                                                #
# ------------------------------------------------------------- #


def print_resources(response: Response, filters: Optional[str] = None):
    resources = response.json()["services"]
    if filters:
        resources = fmt.filter_list(resources, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "TYPE",
            "PRIVATE IP",
            "PUBLIC IP",
        ]
    )
    for res in resources:
        table.row(
            [
                res["id"],
                res["name"],
                res["type"],
                res["local_ip"],
                res["public_ip"],
            ]
        )
    table.print()


@vpc.command("show")
def vpc_show(
    vpc_id: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List resources in network."""
    client = create_client(config, profile)
    response = client.get_services_in_vpc(vpc_id)
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_resources,
    )



"""Manage floating IPs."""

import sys
from typing import Optional, List
from pathlib import Path
from uuid import UUID

import typer
from requests import Response

from twc import fmt
from twc.api import TimewebCloud, ResourceType
from twc.apiwrap import create_client
from twc.typerx import TyperAlias
from .common import (
    verbose_option,
    config_option,
    profile_option,
    yes_option,
    output_format_option,
    load_from_config_callback,
)


floating_ip = TyperAlias(help=__doc__)


def get_floating_ip_id(client: TimewebCloud, ip_addr: str) -> Optional[str]:
    ips = client.get_floating_ips().json()["ips"]
    for ip in ips:
        if ip["ip"] == ip_addr:
            return ip["id"]
    return None


# ------------------------------------------------------------- #
# $ twc floating-ip list                                        #
# ------------------------------------------------------------- #


def _print_floating_ips(response: Response):
    table = fmt.Table()
    table.header(["IP", "PTR", "ZONE", "ANTI_DDOS", "USED_ON"])
    ips = response.json()["ips"]
    for ip in ips:
        used = None
        if ip["resource_type"]:
            used = f"{ip['resource_type']}:{ip['resource_id']}"
        table.row(
            [
                ip["ip"],
                ip["ptr"],
                ip["availability_zone"],
                ip["is_ddos_guard"],
                used,
            ]
        )
    table.print()


@floating_ip.command("list", "ls")
def floating_ip_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """List floating IPs."""
    client = create_client(config, profile)
    response = client.get_floating_ips()
    fmt.printer(
        response,
        output_format=output_format,
        func=_print_floating_ips,
    )


# ------------------------------------------------------------- #
# $ twc floating-ip get                                         #
# ------------------------------------------------------------- #


def _print_floating_ip(response: Response):
    table = fmt.Table()
    table.header(["IP", "PTR", "ZONE", "ANTI_DDOS", "USED_ON"])
    ip = response.json()["ip"]
    used = None
    if ip["resource_type"]:
        used = f"{ip['resource_type']}:{ip['resource_id']}"
    table.row(
        [
            ip["ip"],
            ip["ptr"],
            ip["availability_zone"],
            ip["is_ddos_guard"],
            used,
        ]
    )
    table.print()


@floating_ip.command("get")
def floating_ip_get(
    ip: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get floating IP."""
    client = create_client(config, profile)
    try:
        _ = UUID(ip)
    except ValueError:
        ip = get_floating_ip_id(client, ip)
    response = client.get_floating_ip(ip)
    fmt.printer(
        response,
        output_format=output_format,
        func=_print_floating_ip,
    )


# ------------------------------------------------------------- #
# $ twc floating-ip create                                      #
# ------------------------------------------------------------- #


@floating_ip.command("create")
def floating_ip_create(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    availability_zone: str = typer.Option(
        ...,
        metavar="ZONE",
        envvar="TWC_AVAILABILITY_ZONE",
        show_envvar=False,
        callback=load_from_config_callback,
        help="Availability zone.",
    ),
    ddos_protection: bool = typer.Option(
        False,
        "--ddos-protection",
        show_default=True,
        help="Request IP-address with L3/L4 DDoS protection.",
    ),
):
    """Create new floating IP."""
    client = create_client(config, profile)
    response = client.create_floating_ip(
        availability_zone=availability_zone,
        ddos_protection=ddos_protection,
    )
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["ip"]["ip"]),
    )


# ------------------------------------------------------------- #
# $ twc floating-ip remove                                      #
# ------------------------------------------------------------- #


@floating_ip.command("remove", "rm")
def floating_ip_remove(
    floating_ips: List[str] = typer.Argument(..., metavar="IP..."),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    yes: Optional[bool] = yes_option,
):
    """Remove floating IPs."""
    if not yes:
        typer.confirm("This action cannot be undone, continue?", abort=True)

    client = create_client(config, profile)
    for ip in floating_ips:
        try:
            _ = UUID(ip)
        except ValueError:
            ip = get_floating_ip_id(client, ip)
        response = client.delete_floating_ip(ip)
        if response.status_code == 204:
            print(ip)
        else:
            sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc floating-ip attach                                      #
# ------------------------------------------------------------- #


@floating_ip.command("attach")
def floating_ip_attach(
    ip: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    server: Optional[int] = typer.Option(
        None, help="Attach IP to Cloud Server."
    ),
    balancer: Optional[int] = typer.Option(
        None, help="Attach IP to Load Balancer."
    ),
    database: Optional[int] = typer.Option(
        None, help="Attach IP to managed database cluster."
    ),
):
    """Attach floating IP to service."""
    client = create_client(config, profile)
    try:
        _ = UUID(ip)
    except ValueError:
        ip = get_floating_ip_id(client, ip)
    resource_type = resource_id = None
    if server:
        resource_type = ResourceType.SERVER
        resource_id = server
    if balancer:
        resource_type = ResourceType.BALANCER
        resource_id = balancer
    if database:
        resource_type = ResourceType.DATABASE
        resource_id = database
    response = client.attach_floating_ip(
        ip,
        resource_type=resource_type,
        resource_id=resource_id,
    )
    if not resource_type or not resource_id:
        sys.exit(
            "Error: Please set one of options: ['--server', '--balancer', '--database']"
        )
    if response.status_code == 204:
        print(resource_id)
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc floating-ip detach                                      #
# ------------------------------------------------------------- #


@floating_ip.command("detach")
def floating_ip_detach(
    ip: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
):
    """Detach floating IP from service."""
    client = create_client(config, profile)
    try:
        _ = UUID(ip)
    except ValueError:
        ip = get_floating_ip_id(client, ip)
    response = client.detach_floating_ip(ip)
    if response.status_code == 204:
        print(ip)
    else:
        sys.exit(fmt.printer(response))


# ------------------------------------------------------------- #
# $ twc floating-ip set                                         #
# ------------------------------------------------------------- #


@floating_ip.command("set")
def floating_ip_set(
    ip: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    comment: Optional[str] = typer.Option(None, help="Set comment."),
    ptr: Optional[str] = typer.Option(None, help="Set reverse DNS pointer."),
):
    """Set floating IP parameters."""
    client = create_client(config, profile)
    try:
        _ = UUID(ip)
    except ValueError:
        ip = get_floating_ip_id(client, ip)
    response = client.update_floating_ip(ip, comment=comment, ptr=ptr)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["ip"]["ip"]),
    )



"""Manage Timeweb Cloud account."""

from typing import Optional
from pathlib import Path
from textwrap import dedent

import typer
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.apiwrap import create_client
from .common import (
    verbose_option,
    config_option,
    profile_option,
    output_format_option,
)


whoami = TyperAlias(help="Display current login.", no_args_is_help=False)


# ------------------------------------------------------------- #
# $ twc whoami                                                  #
# ------------------------------------------------------------- #


@whoami.callback(invoke_without_command=True)
def whoami_callback(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Display current login."""
    client = create_client(config, profile)
    response = client.get_account_status()
    login = response.json()["status"]["login"]
    fmt.printer(
        {"login": login, "profile": profile},
        output_format=output_format,
        func=lambda data: print(data["login"]),
    )


# ------------------------------------------------------------- #


account = TyperAlias(help=__doc__)
account_access = TyperAlias(help="Manage account access restrictions.")
account.add_typer(account_access, name="access")


# ------------------------------------------------------------- #
# $ twc account status                                          #
# ------------------------------------------------------------- #


def print_account_status(response: Response):
    """Print table with account info."""
    status = response.json()["status"]
    output = dedent(
        f"""
        Login: {status["login"]}
        Provider: {status["company_info"]["name"]}
        Yandex.Metrika ID: {status["ym_client_id"]}
        Blocked: {status["is_blocked"]}
        Permanently blocked: {status["is_permanent_blocked"]}
        Send bill emails: {status["is_send_bill_letters"]}
        Password changed at: {status["last_password_changed_at"]}
        """
    ).strip()
    print(output)


@account.command("status")
def account_status(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Display account status."""
    client = create_client(config, profile)
    response = client.get_account_status()
    fmt.printer(
        response, output_format=output_format, func=print_account_status
    )


# ------------------------------------------------------------- #
# $ twc account finances                                        #
# ------------------------------------------------------------- #


def print_account_finances(response: Response):
    """Print table with finances info."""
    finances = response.json()["finances"]
    output = dedent(
        f"""
        Balance: {finances["balance"]}
        Currency: {finances["currency"]}
        Discount: {finances["discount_percent"]}
          Discount ends at: {finances["discount_end_date_at"]}
        Hourly cost: {finances["hourly_cost"]}
        Hourly fee: {finances["hourly_fee"]}
        Hours left: {finances["hours_left"]}
        Monthly cost: {finances["monthly_cost"]}
        Monthly fee: {finances["monthly_fee"]}
        Total paid: {finances["total_paid"]}
        Autopay card: {finances["autopay_card_info"]}
        """
    ).strip()
    print(output)


@account.command("finances")
def account_finances(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get finances."""
    client = create_client(config, profile)
    response = client.get_account_finances()
    fmt.printer(
        response, output_format=output_format, func=print_account_finances
    )


# ------------------------------------------------------------- #
# $ twc account access restrictions                             #
# ------------------------------------------------------------- #


def print_restrictions_status(
    response: Response, by_ip: bool, by_country: bool
):
    """Print restrictions info."""
    restrictions = response.json()

    if not by_ip and not by_country:
        by_ip = by_country = True

    if by_ip:
        if restrictions["is_ip_restrictions_enabled"]:
            print("IP restrictions: enabled")
            print("Allowed IPs:")
            for ip_addr in restrictions["white_list"]["ips"]:
                print(f"  {ip_addr}")
        else:
            print("IP restrictions: disabled")

    if by_country:
        if restrictions["is_country_restrictions_enabled"]:
            print("Country restrictions: enabled")
            print("Allowed countries:")
            for country in restrictions["white_list"]["countries"]:
                print(f"  {country}")
        else:
            print("Country restrictions: disabled")


@account_access.command("restrictions")
def restrictions_status(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    by_ip: Optional[bool] = typer.Option(
        False, "--by-ip", help="Display IP restrictions."
    ),
    by_country: Optional[bool] = typer.Option(
        False, "--by-country", help="Display country restrictions."
    ),
):
    """View access restrictions status."""
    client = create_client(config, profile)
    response = client.get_account_restrictions()
    fmt.printer(
        response,
        output_format=output_format,
        func=print_restrictions_status,
        by_ip=by_ip,
        by_country=by_country,
    )



"""Manage apps."""

from typing import Optional
from pathlib import Path

import typer
from requests import Response

from twc import fmt
from twc.typerx import TyperAlias
from twc.apiwrap import create_client
from .common import (
    verbose_option,
    config_option,
    profile_option,
    filter_option,
    output_format_option,
)


apps = TyperAlias(help=__doc__)


# ------------------------------------------------------------- #
# $ twc apps list                                           #
# ------------------------------------------------------------- #


def print_apps(response: Response, filters: Optional[str]):
    """Print table with apps list."""
    # pylint: disable=invalid-name
    _apps = response.json()["apps"]
    if filters:
        _apps = fmt.filter_list(_apps, filters)
    table = fmt.Table()
    table.header(
        [
            "ID",
            "NAME",
            "STATUS",
            "TYPE",
            "IPV4",
        ]
    )
    for app in _apps:
        table.row(
            [
                app["id"],
                app["name"],
                app["status"],
                app["type"],
                app["ip"],
            ]
        )
    table.print()


@apps.command("list", "ls")
def apps_list(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    filters: Optional[str] = filter_option,
):
    """List apps."""
    client = create_client(config, profile)
    response = client.get_apps()
    fmt.printer(
        response,
        output_format=output_format,
        filters=filters,
        func=print_apps,
    )


@apps.command("create")
def app_create(
    yml_config_path: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    status: Optional[bool] = typer.Option(
        False,
        "--status",
        help="Display status and exit with 0 if status is 'started'.",
    ),
):
    """Create app"""
    client = create_client(config, profile)
    response = client.create_app(yml_config_path)
    fmt.printer(
        response,
        output_format=output_format,
        func=lambda response: print(response.json()["app"]["id"]),
    )


def print_vcs_providers(response: Response):
    """Print table with vcs list."""
    # pylint: disable=invalid-name
    providers = response.json()["providers"]
    table = fmt.Table()
    table.header(
        [
            "LOGIN",
            "PROVIDER",
            "PROVIDER_ID",
        ]
    )
    for provider in providers:
        table.row(
            [
                provider["login"],
                provider["provider"],
                provider["provider_id"],
            ]
        )
    table.print()


@apps.command("get-vcs-providers")
def app_get_vcs_providers(
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get VCS providers."""
    client = create_client(config, profile)
    response = client.get_vcs_providers()
    fmt.printer(
        response, output_format=output_format, func=print_vcs_providers
    )


def print_vcs_repositories(response: Response):
    """Print table with vcs list."""
    # pylint: disable=invalid-name
    providers = response.json()["repositories"]
    table = fmt.Table()
    table.header(["FULL NAME", "ID", "NAME", "URL", "IS PRIVATE"])
    for provider in providers:
        table.row(
            [
                provider["full_name"],
                provider["id"],
                provider["name"],
                provider["url"],
                provider["is_private"],
            ]
        )
    table.print()


@apps.command("get-repositories")
def app_get_repositories(
    vcs_provider_id: str,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get repositories."""
    client = create_client(config, profile)
    response = client.get_repositories(vcs_provider_id)
    fmt.printer(
        response, output_format=output_format, func=print_vcs_repositories
    )


def print_apps_tarifs(response: Response, typ: str):
    """Print Timeweb Cloud Apps tarifs."""
    # pylint: disable=invalid-name
    if typ == "backend":
        key = "backend_presets"
    elif typ == "frontend":
        key = "frontend_presets"
    else:
        raise KeyError("Tarifs can be only backend or frontend")

    providers = response.json()[key]
    table = fmt.Table()
    if key == "backend_presets":
        table.header(
            [
                "CPU",
                "CPU FREQUENCY",
                "DESCRIPTION",
                "DESCRIPTION SHORT",
                "DISK",
                "ID",
                "LOCATION",
                "PRICE",
                "RAM",
            ]
        )
        for provider in providers:
            table.row(
                [
                    provider["cpu"],
                    provider["cpu_frequency"],
                    provider["description"],
                    provider["description_short"],
                    provider["disk"],
                    provider["id"],
                    provider["location"],
                    provider["price"],
                    provider["ram"],
                ]
            )
    elif key == "frontend_presets":
        table.header(
            [
                "DESCRIPTION",
                "DESCRIPTION_SHORT",
                "DISK",
                "ID",
                "LOCATION",
                "PRICE",
                "REQUESTS",
            ]
        )
        for provider in providers:
            table.row(
                [
                    provider["description"],
                    provider["description_short"],
                    provider["disk"],
                    provider["id"],
                    provider["location"],
                    provider["price"],
                    provider["requests"],
                ]
            )
    table.print()


@apps.command("list-presets")
def get_apps_tarifs(
    _type: str = typer.Argument(..., metavar="TYPE"),
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Get tarifs; backend or frontend"""
    client = create_client(config, profile)
    response = client.get_apps_tarifs()
    fmt.printer(
        response,
        output_format=output_format,
        typ=_type,
        func=print_apps_tarifs,
    )


def print_app_delete_response(response: Response, app_id: int):
    """Print table with apps list."""
    # pylint: disable=invalid-name
    table = fmt.Table()
    table.header(
        [
            "ID",
        ]
    )
    table.row(
        [
            app_id,
        ]
    )
    table.print()


@apps.command("delete")
def delete_app(
    app_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
):
    """Delete apps."""
    client = create_client(config, profile)
    response = client.delete_app(app_id)
    fmt.printer(
        response,
        output_format=output_format,
        app_id=app_id,
        func=print_app_delete_response,
    )


def get_app(response: Response):
    """Print table with apps list."""
    # pylint: disable=invalid-name
    app = response.json()["app"]
    table = fmt.Table()
    table.header(
        [
            "ID",
            "LOCATION",
            "STATUS",
            "TYPE",
            "IPV4",
        ]
    )
    table.row(
        [
            app["id"],
            app["location"],
            app["status"],
            app["type"],
            app["ip"],
        ]
    )
    table.print()


@apps.command("get")
def app_get(
    app_id: int,
    verbose: Optional[bool] = verbose_option,
    config: Optional[Path] = config_option,
    profile: Optional[str] = profile_option,
    output_format: Optional[str] = output_format_option,
    status: Optional[bool] = typer.Option(
        False,
        "--status",
        help="Display status and exit with 0 if status is 'started'.",
    ),
):
    """Get database info."""
    client = create_client(config, profile)
    response = client.get_app(app_id)
    fmt.printer(response, output_format=output_format, func=get_app)
